import math
import synapse.exc as s_exc
import synapse.common as s_common
import synapse.datamodel as s_datamodel

import synapse.lib.time as s_time
import synapse.lib.const as s_const

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist


class TypesTest(s_t_utils.SynTest):

    def test_type(self):
        # Base type tests, mainly sad paths
        model = s_datamodel.Model()
        t = model.type('bool')
        self.eq(t.info.get('bases'), ())
        self.none(t.getCompOffs('newp'))
        self.raises(s_exc.NoSuchCmpr, t.cmpr, val1=1, name='newp', val2=0)

    def test_hugenum(self):

        model = s_datamodel.Model()
        huge = model.type('hugenum')

        with self.raises(s_exc.BadTypeValu):
            huge.norm('730750818665451459101843')

        with self.raises(s_exc.BadTypeValu):
            huge.norm('-730750818665451459101843')

        with self.raises(s_exc.BadTypeValu):
            huge.norm(None)

        with self.raises(s_exc.BadTypeValu):
            huge.norm('foo')

        self.eq('0.000000000000000000000001', huge.norm('1E-24')[0])
        self.eq('0.000000000000000000000001', huge.norm('1.0E-24')[0])
        self.eq('0.000000000000000000000001', huge.norm('0.000000000000000000000001')[0])

        self.eq('0', huge.norm('1E-25')[0])
        self.eq('0', huge.norm('5E-25')[0])
        self.eq('0.000000000000000000000001', huge.norm('6E-25')[0])
        self.eq('1.000000000000000000000002', huge.norm('1.0000000000000000000000015')[0])

        bign = '730750818665451459101841.000000000000000000000002'
        self.eq(bign, huge.norm(bign)[0])

        big2 = '730750818665451459101841.0000000000000000000000015'
        self.eq(bign, huge.norm(big2)[0])

        bign = '-730750818665451459101841.000000000000000000000002'
        self.eq(bign, huge.norm(bign)[0])

        big2 = '-730750818665451459101841.0000000000000000000000015'
        self.eq(bign, huge.norm(big2)[0])

    def test_taxonomy(self):

        model = s_datamodel.Model()
        taxo = model.type('taxonomy')
        self.eq('foo.bar.baz.', taxo.norm('foo.bar.baz')[0])
        self.eq('foo.bar.baz.', taxo.norm('foo.bar.baz.')[0])
        self.eq('foo.bar.baz.', taxo.norm('foo.bar.baz.')[0])
        self.eq('foo.bar.baz.', taxo.norm(('foo', 'bar', 'baz'))[0])
        self.eq('foo.b_a_r.baz.', taxo.norm('foo.b-a-r.baz.')[0])
        self.eq('foo.b_a_r.baz.', taxo.norm('foo.  b   a   r  .baz.')[0])

        with self.raises(s_exc.BadTypeValu):
            taxo.norm('foo.---.baz')

        norm, info = taxo.norm('foo.bar.baz')
        self.eq(2, info['subs']['depth'])
        self.eq('baz', info['subs']['base'])
        self.eq('foo.bar.', info['subs']['parent'])

    def test_duration(self):
        model = s_datamodel.Model()
        t = model.type('duration')

        self.eq('2D 00:00:00.000', t.repr(172800000))
        self.eq('00:05:00.333', t.repr(300333))
        self.eq('11D 11:47:12.344', t.repr(992832344))

        self.eq(300333, t.norm('00:05:00.333')[0])
        self.eq(992832344, t.norm('11D 11:47:12.344')[0])

        self.eq(172800000, t.norm('2D')[0])
        self.eq(60000, t.norm('1:00')[0])
        self.eq(60200, t.norm('1:00.2')[0])
        self.eq(9999, t.norm('9.9999')[0])

        with self.raises(s_exc.BadTypeValu):
            t.norm('    ')

        with self.raises(s_exc.BadTypeValu):
            t.norm('1:2:3:4')

        with self.raises(s_exc.BadTypeValu):
            t.norm('1:a:b')

    def test_bool(self):
        model = s_datamodel.Model()
        t = model.type('bool')

        self.eq(t.norm(-1), (1, {}))
        self.eq(t.norm(0), (0, {}))
        self.eq(t.norm(1), (1, {}))
        self.eq(t.norm(2), (1, {}))
        self.eq(t.norm(True), (1, {}))
        self.eq(t.norm(False), (0, {}))

        self.eq(t.norm('-1'), (1, {}))
        self.eq(t.norm('0'), (0, {}))
        self.eq(t.norm('1'), (1, {}))

        for s in ('trUe', 'T', 'y', ' YES', 'On '):
            self.eq(t.norm(s), (1, {}))

        for s in ('faLSe', 'F', 'n', 'NO', 'Off '):
            self.eq(t.norm(s), (0, {}))

        self.raises(s_exc.BadTypeValu, t.norm, 'a')

        self.eq(t.repr(1), 'True')
        self.eq(t.repr(0), 'False')

    async def test_comp(self):
        async with self.getTestCore() as core:
            t = 'test:complexcomp'
            valu = ('123', 'HAHA')
            async with await core.snap() as snap:
                node = await snap.addNode(t, valu)
            pnode = node.pack(dorepr=True)
            self.eq(pnode[0], (t, (123, 'haha')))
            self.eq(pnode[1].get('repr'), ('123', 'haha'))
            self.eq(pnode[1].get('reprs').get('foo'), '123')
            self.notin('bar', pnode[1].get('reprs'))
            self.eq(node.get('foo'), 123)
            self.eq(node.get('bar'), 'haha')

            typ = core.model.type(t)
            self.eq(typ.info.get('bases'), ('comp',))
            self.raises(s_exc.BadTypeValu, typ.norm,
                        (123, 'haha', 'newp'))
            self.eq(0, typ.getCompOffs('foo'))
            self.eq(1, typ.getCompOffs('bar'))
            self.none(typ.getCompOffs('newp'))

    def test_guid(self):
        model = s_datamodel.Model()

        guid = 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
        self.eq(guid.lower(), model.type('guid').norm(guid)[0])
        self.raises(s_exc.BadTypeValu, model.type('guid').norm, 'visi')

        guid = model.type('guid').norm('*')[0]
        self.true(s_common.isguid(guid))

        objs = [1, 2, 'three', {'four': 5}]
        tobjs = tuple(objs)
        lnorm, _ = model.type('guid').norm(objs)
        tnorm, _ = model.type('guid').norm(tobjs)
        self.true(s_common.isguid(lnorm))
        self.eq(lnorm, tnorm)

    async def test_hex(self):

        async with self.getTestCore() as core:

            # Bad configurations are not allowed for the type
            with self.raises(s_exc.BadConfValu):
                core.model.type('hex').clone({'size': 1})

            with self.raises(s_exc.BadConfValu):
                core.model.type('hex').clone({'size': -1})

            t = core.model.type('test:hexa')
            # Test norming to index values
            testvectors = [
                ('0C', b'\x0c'),
                ('0X010001', b'\x01\x00\x01'),
                ('0FfF', b'\x0f\xff'),
                ('f12A3e', b'\xf1\x2a\x3e'),
                (b'\x01\x00\x01', b'\x01\x00\x01'),
                (b'\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~',
                 b'\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~'),
                (65537, s_exc.BadTypeValu),
            ]

            for v, b in testvectors:
                if isinstance(b, bytes):
                    r, subs = t.norm(v)
                    self.isinstance(r, str)
                    self.eq(subs, {})
                else:
                    self.raises(b, t.norm, v)

            # width = 4
            testvectors4 = [
                ('d41d', b'\xd4\x1d'),
                (b'\x10\x01', b'\x10\x01'),
                ('01', s_exc.BadTypeValu),
                ('010101', s_exc.BadTypeValu),
                (b'\x10\x01\xff', s_exc.BadTypeValu),
                (b'\xff', s_exc.BadTypeValu),
                ('01\udcfe0101', s_exc.BadTypeValu),
            ]
            t = core.model.type('test:hex4')
            for v, b in testvectors4:
                if isinstance(b, bytes):
                    r, subs = t.norm(v)
                    self.isinstance(r, str)
                    self.eq(subs, {})
                else:
                    self.raises(b, t.norm, v)

            # Do some node creation and lifting
            async with await core.snap() as snap:
                node = await snap.addNode('test:hexa', '010001')
                self.eq(node.ndef[1], '010001')

            async with await core.snap() as snap:
                nodes = await snap.nodes('test:hexa=010001')
                self.len(1, nodes)

                nodes = await snap.nodes('test:hexa=$byts', opts={'vars': {'byts': b'\x01\x00\x01'}})
                self.len(1, nodes)

            # Do some fancy prefix searches for test:hexa
            valus = ['deadb33f',
                     'deadb33fb33f',
                     'deadb3b3',
                     'deaddead',
                     'DEADBEEF']
            async with await core.snap() as snap:
                for valu in valus:
                    node = await snap.addNode('test:hexa', valu)

            async with await core.snap() as snap:
                nodes = await snap.nodes('test:hexa=dead*')
                self.len(5, nodes)

                nodes = await snap.nodes('test:hexa=deadb3*')
                self.len(3, nodes)

                nodes = await snap.nodes('test:hexa=deadb33fb3*')
                self.len(1, nodes)

                nodes = await snap.nodes('test:hexa=deadde*')
                self.len(1, nodes)

                nodes = await snap.nodes('test:hexa=b33f*')
                self.len(0, nodes)

            # Do some fancy prefix searches for test:hex4
            valus = ['0000',
                     '0100',
                     '01ff',
                     '0200',
                     ]
            async with await core.snap() as snap:
                for valu in valus:
                    node = await snap.addNode('test:hex4', valu)

            async with await core.snap() as snap:
                nodes = await snap.nodes('test:hex4=00*')
                self.len(1, nodes)

                nodes = await snap.nodes('test:hex4=01*')
                self.len(2, nodes)

                nodes = await snap.nodes('test:hex4=02*')
                self.len(1, nodes)

                # You can ask for a longer prefix then allowed
                # but you'll get no results
                nodes = await snap.nodes('test:hex4=022020*')
                self.len(0, nodes)

            self.len(1, await core.nodes('[test:hexa=0xf00fb33b00000000]'))
            self.len(1, await core.nodes('test:hexa=0xf00fb33b00000000'))
            self.len(1, await core.nodes('test:hexa^=0xf00fb33b'))

    def test_int(self):

        model = s_datamodel.Model()
        t = model.type('int')

        # test ranges
        self.nn(t.norm(-2**63))
        self.raises(s_exc.BadTypeValu, t.norm, (-2**63) - 1)
        self.nn(t.norm(2**63 - 1))
        self.raises(s_exc.BadTypeValu, t.norm, 2**63)

        # test base types that Model snaps in...
        self.eq(t.norm('100')[0], 100)
        self.eq(t.norm('0x20')[0], 32)
        self.raises(s_exc.BadTypeValu, t.norm, 'newp')
        self.eq(t.norm(True)[0], 1)
        self.eq(t.norm(False)[0], 0)

        # Test merge
        self.eq(30, t.merge(20, 30))
        self.eq(20, t.merge(30, 20))
        self.eq(20, t.merge(20, 20))

        # Test min and max
        minmax = model.type('int').clone({'min': 10, 'max': 30})
        self.eq(20, minmax.norm(20)[0])
        self.raises(s_exc.BadTypeValu, minmax.norm, 9)
        self.raises(s_exc.BadTypeValu, minmax.norm, 31)
        ismin = model.type('int').clone({'ismin': True})
        self.eq(20, ismin.merge(20, 30))
        ismin = model.type('int').clone({'ismax': True})
        self.eq(30, ismin.merge(20, 30))

        # Test unsigned
        uint64 = model.type('int').clone({'signed': False})
        self.eq(uint64.norm(0)[0], 0)
        self.eq(uint64.norm(-0)[0], 0)
        self.raises(s_exc.BadTypeValu, uint64.norm, -1)

        # Test size, 8bit signed
        int8 = model.type('int').clone({'size': 1})
        self.eq(int8.norm(127)[0], 127)
        self.eq(int8.norm(0)[0], 0)
        self.eq(int8.norm(-128)[0], -128)
        self.raises(s_exc.BadTypeValu, int8.norm, 128)
        self.raises(s_exc.BadTypeValu, int8.norm, -129)

        # Test size, 128bit signed
        int128 = model.type('int').clone({'size': 16})
        self.eq(int128.norm(2**127 - 1)[0], 170141183460469231731687303715884105727)
        self.eq(int128.norm(0)[0], 0)
        self.eq(int128.norm(-2**127)[0], -170141183460469231731687303715884105728)
        self.raises(s_exc.BadTypeValu, int128.norm, 170141183460469231731687303715884105728)
        self.raises(s_exc.BadTypeValu, int128.norm, -170141183460469231731687303715884105729)

        # test both unsigned and signed comparators
        self.true(uint64.cmpr(10, '<', 20))
        self.true(uint64.cmpr(10, '<=', 20))
        self.true(uint64.cmpr(20, '<=', 20))

        self.true(uint64.cmpr(20, '>', 10))
        self.true(uint64.cmpr(20, '>=', 10))
        self.true(uint64.cmpr(20, '>=', 20))

        self.true(int8.cmpr(-10, '<', 20))
        self.true(int8.cmpr(-10, '<=', 20))
        self.true(int8.cmpr(-20, '<=', -20))

        self.true(int8.cmpr(20, '>', -10))
        self.true(int8.cmpr(20, '>=', -10))
        self.true(int8.cmpr(-20, '>=', -20))

        # test integer enums for repr and norm
        eint = model.type('int').clone({'enums': ((1, 'hehe'), (2, 'haha'))})

        self.eq(1, eint.norm(1)[0])
        self.eq(1, eint.norm('1')[0])
        self.eq(1, eint.norm('hehe')[0])
        self.eq(2, eint.norm('haha')[0])
        self.eq(2, eint.norm('HAHA')[0])

        self.eq('hehe', eint.repr(1))
        self.eq('haha', eint.repr(2))

        self.raises(s_exc.BadTypeValu, eint.norm, 0)
        self.raises(s_exc.BadTypeValu, eint.norm, '0')
        self.raises(s_exc.BadTypeValu, eint.norm, 'newp')

        # Invalid Config
        self.raises(s_exc.BadTypeDef, model.type('int').clone, {'min': 100, 'max': 1})
        self.raises(s_exc.BadTypeDef, model.type('int').clone, {'enums': ((1, 'hehe'), (2, 'haha'), (3, 'HAHA'))})
        self.raises(s_exc.BadTypeDef, model.type('int').clone, {'enums': ((1, 'hehe'), (2, 'haha'), (2, 'beep'))})

    async def test_float(self):
        model = s_datamodel.Model()
        t = model.type('float')

        self.nn(t.norm(1.2345)[0])
        self.eq(t.norm('inf')[0], math.inf)
        self.eq(t.norm('-inf')[0], -math.inf)
        self.true(math.isnan(t.norm('NaN')[0]))
        self.eq(t.norm('-0.0')[0], -0.0)
        self.eq(t.norm('42')[0], 42.0)
        minmax = model.type('float').clone({'min': -10.0, 'max': 100.0, 'maxisvalid': True, 'minisvalid': False})
        self.raises(s_exc.BadTypeValu, minmax.norm, 'NaN')
        self.raises(s_exc.BadTypeValu, minmax.norm, '-inf')
        self.raises(s_exc.BadTypeValu, minmax.norm, 'inf')
        self.raises(s_exc.BadTypeValu, minmax.norm, '-10')
        self.raises(s_exc.BadTypeValu, minmax.norm, '-10.00001')
        self.raises(s_exc.BadTypeValu, minmax.norm, '100.00001')
        self.eq(minmax.norm('100.000')[0], 100.0)
        self.true(t.cmpr(10, '<', 20.0))
        self.false(t.cmpr(-math.inf, '>', math.inf))
        self.false(t.cmpr(-math.nan, '<=', math.inf))
        self.true(t.cmpr('inf', '>=', '-0.0'))

        async with self.getTestCore() as core:
            nodes = await core.nodes('[ test:float=42.0 ]')
            self.len(1, nodes)
            nodes = await core.nodes('[ test:float=inf ]')
            self.len(1, nodes)

            self.len(1, await core.nodes('[ test:float=42.0 :closed=0.0]'))
            await self.asyncraises(s_exc.BadTypeValu, core.nodes('[ test:float=42.0 :closed=-1.0]'))
            self.len(1, await core.nodes('[ test:float=42.0 :closed=360.0]'))
            await self.asyncraises(s_exc.BadTypeValu, core.nodes('[ test:float=42.0 :closed=NaN]'))
            await self.asyncraises(s_exc.BadTypeValu, core.nodes('[ test:float=42.0 :closed=360.1]'))

            await self.asyncraises(s_exc.BadTypeValu, core.nodes('[ test:float=42.0 :open=0.0]'))
            await self.asyncraises(s_exc.BadTypeValu, core.nodes('[ test:float=42.0 :open=360.0]'))
            self.len(1, await core.nodes('[ test:float=42.0 :open=0.001]'))
            self.len(1, await core.nodes('[ test:float=42.0 :open=359.0]'))

    async def test_ival(self):
        model = s_datamodel.Model()
        ival = model.types.get('ival')

        self.eq(('2016/01/01 00:00:00.000', '2017/01/01 00:00:00.000'), ival.repr(ival.norm(('2016', '2017'))[0]))

        self.gt(s_common.now(), ival._normRelStr('-1 min'))

        self.eq((0, 5356800000), ival.norm((0, '1970-03-04'))[0])
        self.eq((1451606400000, 1451606400001), ival.norm('2016')[0])
        self.eq((1451606400000, 1451606400001), ival.norm(1451606400000)[0])
        self.eq((1451606400000, 1451606400001), ival.norm('2016')[0])
        self.eq((1451606400000, 1483228800000), ival.norm(('2016', '  2017'))[0])
        self.eq((1451606400000, 1483228800000), ival.norm(('2016-01-01', '  2017'))[0])
        self.eq((1451606400000, 1483142400000), ival.norm(('2016', '+365 days'))[0])
        self.eq((1448150400000, 1451606400000), ival.norm(('2016', '-40 days'))[0])
        self.eq((1447891200000, 1451347200000), ival.norm(('2016-3days', '-40 days   '))[0])
        self.eq((1451347200000, 0x7fffffffffffffff), ival.norm(('2016-3days', '?'))[0])
        self.eq((1593576000000, 1593576000001), ival.norm('2020-07-04:00')[0])
        self.eq((1594124993000, 1594124993001), ival.norm('2020-07-07T16:29:53+04:00')[0])
        self.eq((1594153793000, 1594153793001), ival.norm('2020-07-07T16:29:53-04:00')[0])
        self.eq((1594211393000, 1594211393001), ival.norm('20200707162953+04:00+1day')[0])
        self.eq((1594038593000, 1594038593001), ival.norm('20200707162953+04:00-1day')[0])
        self.eq((1594240193000, 1594240193001), ival.norm('20200707162953-04:00+1day')[0])
        self.eq((1594067393000, 1594067393001), ival.norm('20200707162953-04:00-1day')[0])

        start = s_common.now() + s_time.oneday - 1
        end = ival.norm(('now', '+1day'))[0][1]
        self.lt(start, end)

        oldv = ival.norm(('2016', '2017'))[0]
        newv = ival.norm(('2015', '2018'))[0]
        self.eq((1420070400000, 1514764800000), ival.merge(oldv, newv))

        self.eq((1420070400000, 1420070400001), ival.norm(('2015', '2015'))[0])

        self.raises(s_exc.BadTypeValu, ival.norm, '?')
        self.raises(s_exc.BadTypeValu, ival.norm, ('', ''))
        self.raises(s_exc.BadTypeValu, ival.norm, ('2016-3days', '+77days', '-40days'))
        self.raises(s_exc.BadTypeValu, ival.norm, ('?', '-1 day'))

        async with self.getTestCore() as core:

            t = core.model.type('test:time')

            async with await core.snap() as snap:
                node = await snap.addNode('test:str', 'a', {'tick': '2014', '.seen': ('2005', '2006')})
                await node.addTag('foo', valu=('2000', '2001'))

                node = await snap.addNode('test:str', 'b', {'tick': '2015', '.seen': ('8679', '9000')})
                await node.addTag('foo', valu=('2015', '2018'))

                node = await snap.addNode('test:str', 'c', {'tick': '2016', '.seen': ('now-5days', 'now-1day')})
                await node.addTag('bar', valu=('1970', '1990'))

                node = await snap.addNode('test:str', 'd', {'tick': 'now', '.seen': ('now-10days', '?')})
                await node.addTag('baz', valu='now')

                node = await snap.addNode('test:str', 'e', {'tick': 'now-3days', '.seen': ('now+1day', 'now+5days')})
                await node.addTag('biz', valu=('now-1day', 'now+1day'))

                # node whose primary prop is an ival
                node = await snap.addNode('test:ival', (0, 10), {'interval': ("now", "now+4days")})
                node = await snap.addNode('test:ival', (50, 100), {'interval': ("now-2days", "now+2days")})
                node = await snap.addNode('test:ival', ("1995", "1997"), {'interval': ("2010", "2011")})
                node = await snap.addNode('test:ival', ("now-2days", "now+4days"), {'interval': ("201006", "20100605")})
                node = await snap.addNode('test:ival', ("now+21days", "?"), {'interval': ("2000", "2001")})

                # tag of tags
                node = (await snap.nodes('syn:tag=foo'))[0]
                await node.addTag('v.p', valu=('2005', '2006'))

                node = (await snap.nodes('syn:tag=bar'))[0]
                await node.addTag('vert.proj', valu=('20110605', 'now'))

                node = (await snap.nodes('syn:tag=biz'))[0]
                await node.addTag('vertex.project', valu=('now-5days', 'now'))

            await self.agenraises(s_exc.BadSyntax, core.eval('test:str :tick=(20150102, "-4 day")'))

            await self.agenlen(1, core.eval('test:str +:tick@=(now, "-1 day")'))
            await self.agenlen(1, core.eval('test:str +:tick@=2015'))
            await self.agenlen(1, core.eval('test:str +:tick@=(2015, "+1 day")'))
            await self.agenlen(1, core.eval('test:str +:tick@=(20150102+1day, "-4 day")'))
            await self.agenlen(1, core.eval('test:str +:tick@=(20150102, "-4 day")'))
            await self.agenlen(1, core.eval('test:str +:tick@=(now, "-1 day")'))
            await self.agenlen(1, core.eval('test:str +:tick@=("now-1day", "?")'))
            await self.agenlen(1, core.eval('test:str +:tick@=("now+2days", "-3 day")'))
            await self.agenlen(0, core.eval('test:str +:tick@=("now", "now+3days")'))
            await self.agenlen(1, core.eval('test:str +:tick@=("now-2days","now")'))
            await self.agenlen(0, core.eval('test:str +:tick@=("2011", "2014")'))
            await self.agenlen(1, core.eval('test:str +:tick@=("2014", "20140601")'))

            await self.agenlen(1, core.eval('test:str:tick@=(now, "-1 day")'))
            await self.agenlen(1, core.eval('test:str:tick@=2015'))
            await self.agenlen(1, core.eval('test:str:tick@=(2015, "+1 day")'))
            await self.agenlen(1, core.eval('test:str:tick@=(20150102+1day, "-4 day")'))
            await self.agenlen(1, core.eval('test:str:tick@=(20150102, "-4 day")'))
            await self.agenlen(1, core.eval('test:str:tick@=(now, "-1 day")'))
            await self.agenlen(1, core.eval('test:str:tick@=("now-1day", "?")'))
            await self.agenlen(1, core.eval('test:str:tick@=("now+2days", "-3 day")'))
            await self.agenlen(0, core.eval('test:str:tick@=("now", "now+3days")'))
            await self.agenlen(1, core.eval('test:str:tick@=("now-2days","now")'))
            await self.agenlen(0, core.eval('test:str:tick@=("2011", "2014")'))
            await self.agenlen(1, core.eval('test:str:tick@=("2014", "20140601")'))

            await self.agenlen(0, core.eval('.seen@=("2004", "2005")'))
            await self.agenlen(1, core.eval('.seen@=("9000", "9001")'))

            await self.agenlen(2, core.eval('.seen@=("now+6days", "?")'))
            await self.agenlen(2, core.eval('.seen@=(now, "-4 days")'))
            await self.agenlen(2, core.eval('.seen@=(8900, 9500)'))
            await self.agenlen(1, core.eval('.seen@=("2004", "20050201")'))
            await self.agenlen(2, core.eval('.seen@=("now", "-3 days")'))

            await self.agenlen(1, core.eval('test:ival@=1970'))
            await self.agenlen(5, core.eval('test:ival@=(1970, "now+100days")'))
            await self.agenlen(1, core.eval('test:ival@="now"'))
            await self.agenlen(1, core.eval('test:ival@=("now+1day", "now+6days")'))
            await self.agenlen(1, core.eval('test:ival@=("now-9days", "now-1day")'))
            await self.agenlen(1, core.eval('test:ival@=("now-3days", "now+3days")'))
            await self.agenlen(0, core.eval('test:ival@=("1993", "1995")'))
            await self.agenlen(0, core.eval('test:ival@=("1997", "1998")'))
            await self.agenlen(1, core.eval('test:ival=("1995", "1997")'))

            await self.agenlen(1, core.eval('test:ival:interval@="now+2days"'))
            await self.agenlen(0, core.eval('test:ival:interval@=("now-4days","now-3days")'))
            await self.agenlen(0, core.eval('test:ival:interval@=("now+4days","now+6days")'))
            await self.agenlen(1, core.eval('test:ival:interval@=("now-3days","now-1days")'))
            await self.agenlen(1, core.eval('test:ival:interval@=("now+3days","now+6days")'))
            await self.agenlen(2, core.eval('test:ival:interval@="now+1day"'))
            await self.agenlen(2, core.eval('test:ival:interval@=("20100602","20100603")'))
            await self.agenlen(2, core.eval('test:ival:interval@=("now-10days","now+10days")'))
            await self.agenlen(0, core.eval('test:ival:interval@=("1999", "2000")'))
            await self.agenlen(0, core.eval('test:ival:interval@=("2001", "2002")'))

            await self.agenlen(1, core.eval('test:ival +:interval@="now+2days"'))
            await self.agenlen(0, core.eval('test:ival +:interval@=("now-4days","now-3days")'))
            await self.agenlen(0, core.eval('test:ival +:interval@=("now+4days","now+6days")'))
            await self.agenlen(1, core.eval('test:ival +:interval@=("now-3days","now-1days")'))
            await self.agenlen(1, core.eval('test:ival +:interval@=("now+3days","now+6days")'))
            await self.agenlen(2, core.eval('test:ival +:interval@="now+1day"'))
            await self.agenlen(2, core.eval('test:ival +:interval@=("20100602","20100603")'))
            await self.agenlen(2, core.eval('test:ival +:interval@=("now-10days","now+10days")'))
            await self.agenlen(0, core.eval('test:ival +:interval@=("1999", "2000")'))
            await self.agenlen(0, core.eval('test:ival +:interval@=("2001", "2002")'))

            await self.agenlen(0, core.eval('#foo@=("2013", "2015")'))
            await self.agenlen(0, core.eval('#foo@=("2018", "2019")'))
            await self.agenlen(1, core.eval('#foo@=("1999", "2002")'))
            await self.agenlen(1, core.eval('#foo@="2015"'))
            await self.agenlen(1, core.eval('#foo@=("2010", "20150601")'))
            await self.agenlen(2, core.eval('#foo@=("2000", "2017")'))
            await self.agenlen(1, core.eval('#bar@=("1985", "1995")'))
            await self.agenlen(0, core.eval('#bar@="2000"'))
            await self.agenlen(1, core.eval('#baz@=("now","-1 day")'))
            await self.agenlen(1, core.eval('#baz@=("now-1day", "+1day")'))
            await self.agenlen(1, core.eval('#biz@="now"'))

            await self.agenlen(0, core.eval('#foo +#foo@=("2013", "2015")'))
            await self.agenlen(0, core.eval('#foo +#foo@=("2018", "2019")'))
            await self.agenlen(1, core.eval('#foo +#foo@=("1999", "2002")'))
            await self.agenlen(1, core.eval('#foo +#foo@="2015"'))
            await self.agenlen(1, core.eval('#foo +#foo@=("2010", "20150601")'))
            await self.agenlen(2, core.eval('#foo +#foo@=("2000", "2017")'))
            await self.agenlen(1, core.eval('#bar +#bar@=("1985", "1995")'))
            await self.agenlen(0, core.eval('#bar +#bar@="2000"'))
            await self.agenlen(1, core.eval('#baz +#baz@=("now","-1 day")'))
            await self.agenlen(1, core.eval('#baz +#baz@=("now-1day", "+1day")'))
            await self.agenlen(1, core.eval('#biz +#biz@="now"'))

            await self.agenlen(1, core.eval('#foo=("2015", "2018")'))

            await self.agenlen(0, core.eval('test:str#foo@=("2013", "2015")'))
            await self.agenlen(0, core.eval('test:str#foo@=("2018", "2019")'))
            await self.agenlen(1, core.eval('test:str#foo@=("1999", "2002")'))
            await self.agenlen(1, core.eval('test:str#foo@="2015"'))
            await self.agenlen(1, core.eval('test:str#foo@=("2010", "20150601")'))
            await self.agenlen(2, core.eval('test:str#foo@=("2000", "2017")'))
            await self.agenlen(1, core.eval('test:str#bar@=("1985", "1995")'))
            await self.agenlen(0, core.eval('test:str#bar@="2000"'))
            await self.agenlen(1, core.eval('test:str#baz@=("now","-1 day")'))
            await self.agenlen(1, core.eval('test:str#baz@=("now-1day", "+1day")'))
            await self.agenlen(1, core.eval('test:str#biz@="now"'))

            await self.agenlen(0, core.eval('test:str +#foo@=("2013", "2015")'))
            await self.agenlen(0, core.eval('test:str +#foo@=("2018", "2019")'))
            await self.agenlen(1, core.eval('test:str +#foo@=("1999", "2002")'))
            await self.agenlen(1, core.eval('test:str +#foo@="2015"'))
            await self.agenlen(1, core.eval('test:str +#foo@=("2010", "20150601")'))
            await self.agenlen(2, core.eval('test:str +#foo@=("2000", "2017")'))
            await self.agenlen(1, core.eval('test:str +#bar@=("1985", "1995")'))
            await self.agenlen(0, core.eval('test:str +#bar@="2000"'))
            await self.agenlen(1, core.eval('test:str +#baz@=("now","-1 day")'))
            await self.agenlen(1, core.eval('test:str +#baz@=("now-1day", "+1day")'))
            await self.agenlen(1, core.eval('test:str +#biz@="now"'))

            await self.agenlen(0, core.eval('##v.p@=("2003", "2005")'))
            await self.agenlen(0, core.eval('##v.p@=("2006", "2008")'))
            await self.agenlen(1, core.eval('##vert.proj@="2016"'))
            await self.agenlen(1, core.eval('##vert.proj@=("2010", "2012")'))
            await self.agenlen(1, core.eval('##vert.proj@=("2016", "now+6days")'))
            await self.agenlen(1, core.eval('##vert.proj@=("1995", "now+6 days")'))
            await self.agenlen(1, core.eval('##vertex.project@=("now-9days", "now-3days")'))

            await self.agenlen(0, core.eval('test:str +:tick@=(2020, 2000)'))

            now = s_common.now()
            nodes = await alist(core.eval('[test:guid="*" .seen=("-1 day","?")]'))
            node = nodes[0]
            valu = node.get('.seen')
            self.eq(valu[1], ival.futsize)
            self.true(now - s_const.day <= valu[0] < now)

            # Sad Paths
            q = '[test:str=newp .seen=(2018/03/31,2018/03/30)]'
            await self.agenraises(s_exc.BadTypeValu, core.eval(q))
            q = '[test:str=newp .seen=("+-1 day","+-1 day")]'
            await self.agenraises(s_exc.BadTypeValu, core.eval(q))
            q = '[test:str=newp  .seen=("?","?")]'
            await self.agenraises(s_exc.BadTypeValu, core.eval(q))
            q = '[test:str=newp .seen=(2008, 2019, 2000)]'
            await self.agenraises(s_exc.BadTypeValu, core.eval(q))
            q = '[test:str=newp .seen=("?","-1 day")]'
            await self.agenraises(s_exc.BadTypeValu, core.eval(q))
            # *range= not supported for ival
            q = 'test:str +.seen*range=((20090601, 20090701), (20110905, 20110906,))'
            await self.agenraises(s_exc.NoSuchCmpr, core.eval(q))
            q = 'test:str.seen*range=((20090601, 20090701), (20110905, 20110906,))'
            await self.agenraises(s_exc.NoSuchCmpr, core.eval(q))

    async def test_loc(self):
        model = s_datamodel.Model()
        loctype = model.types.get('loc')

        self.eq('us.va', loctype.norm('US.    VA')[0])
        self.eq('', loctype.norm('')[0])
        self.eq('us.va.ओं.reston', loctype.norm('US.    VA.ओं.reston')[0])

        async with self.getTestCore() as core:
            await self.agenlen(1, core.eval('[test:int=1 :loc=us.va.syria]'))
            await self.agenlen(1, core.eval('[test:int=2 :loc=us.va.sydney]'))
            await self.agenlen(1, core.eval('[test:int=3 :loc=""]'))
            await self.agenlen(1, core.eval('[test:int=4 :loc=us.va.fairfax]'))
            await self.agenlen(1, core.eval('[test:int=5 :loc=us.va.fairfax.reston]'))
            await self.agenlen(1, core.eval('[test:int=6 :loc=us.va.fairfax.restonheights]'))
            await self.agenlen(1, core.eval('[test:int=7 :loc=us.va.fairfax.herndon]'))
            await self.agenlen(1, core.eval('[test:int=8 :loc=us.ca.sandiego]'))
            await self.agenlen(1, core.eval('[test:int=9 :loc=us.ओं]'))
            await self.agenlen(1, core.eval('[test:int=10 :loc=us.va]'))
            await self.agenlen(1, core.eval('[test:int=11 :loc=us]'))
            await self.agenlen(1, core.eval('[test:int=12 :loc=us]'))

            await self.agenlen(1, core.eval('test:int:loc=us.va.syria'))
            await self.agenlen(1, core.eval('test:int:loc=us.va.sydney'))
            await self.agenlen(0, core.eval('test:int:loc=us.va.sy'))
            await self.agenlen(1, core.eval('test:int:loc=us.va'))
            await self.agenlen(0, core.eval('test:int:loc=us.v'))
            await self.agenlen(2, core.eval('test:int:loc=us'))
            await self.agenlen(0, core.eval('test:int:loc=u'))
            await self.agenlen(1, core.eval('test:int:loc=""'))

            await self.agenlen(1, core.eval('test:int +:loc="us.va. syria"'))
            await self.agenlen(1, core.eval('test:int +:loc=us.va.sydney'))
            await self.agenlen(0, core.eval('test:int +:loc=us.va.sy'))
            await self.agenlen(1, core.eval('test:int +:loc=us.va'))
            await self.agenlen(0, core.eval('test:int +:loc=us.v'))
            await self.agenlen(2, core.eval('test:int +:loc=us'))
            await self.agenlen(0, core.eval('test:int +:loc=u'))
            await self.agenlen(1, core.eval('test:int +:loc=""'))

            await self.agenlen(0, core.eval('test:int +:loc^=u'))
            await self.agenlen(11, core.eval('test:int +:loc^=us'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.v'))
            await self.agenlen(7, core.eval('test:int +:loc^=us.va'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.va.'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.va.fair'))
            await self.agenlen(4, core.eval('test:int +:loc^=us.va.fairfax'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.va.fairfax.'))
            await self.agenlen(1, core.eval('test:int +:loc^=us.va.fairfax.reston'))
            await self.agenlen(0, core.eval('test:int +:loc^=us.va.fairfax.chantilly'))
            await self.agenlen(1, core.eval('test:int +:loc^=""'))
            await self.agenlen(0, core.eval('test:int +:loc^=23'))

            await self.agenlen(0, core.eval('test:int:loc^=u'))
            await self.agenlen(11, core.eval('test:int:loc^=us'))
            await self.agenlen(0, core.eval('test:int:loc^=us.'))
            await self.agenlen(0, core.eval('test:int:loc^=us.v'))
            await self.agenlen(7, core.eval('test:int:loc^=us.va'))
            await self.agenlen(0, core.eval('test:int:loc^=us.va.'))
            await self.agenlen(0, core.eval('test:int:loc^=us.va.fair'))
            await self.agenlen(4, core.eval('test:int:loc^=us.va.fairfax'))
            await self.agenlen(0, core.eval('test:int:loc^=us.va.fairfax.'))
            await self.agenlen(1, core.eval('test:int:loc^=us.va.fairfax.reston'))
            await self.agenlen(0, core.eval('test:int:loc^=us.va.fairfax.chantilly'))
            await self.agenlen(1, core.eval('test:int:loc^=""'))
            await self.agenlen(0, core.eval('test:int:loc^=23'))

    def test_ndef(self):
        model = s_datamodel.Model()
        model.addDataModels([('test', s_t_utils.testmodel)])
        t = model.type('test:ndef')

        norm, info = t.norm(('test:str', 'Foobar!'))
        self.eq(norm, ('test:str', 'Foobar!'))
        self.eq(info, {'adds': (('test:str', 'Foobar!', {}),),
                       'subs': {'form': 'test:str'}})

        rval = t.repr(('test:str', 'Foobar!'))
        self.eq(rval, ('test:str', 'Foobar!'))
        rval = t.repr(('test:int', 1234))
        self.eq(rval, ('test:int', '1234'))

        self.raises(s_exc.NoSuchForm, t.norm, ('test:newp', 'newp'))
        self.raises(s_exc.NoSuchForm, t.repr, ('test:newp', 'newp'))
        self.raises(s_exc.BadTypeValu, t.norm, ('newp',))

    def test_nodeprop(self):
        model = s_datamodel.Model()
        model.addDataModels([('test', s_t_utils.testmodel)])
        t = model.type('nodeprop')

        expected = (('test:str', 'This is a sTring'), {'subs': {'prop': 'test:str'}})
        self.eq(t.norm('test:str=This is a sTring'), expected)
        self.eq(t.norm(('test:str', 'This is a sTring')), expected)

        self.raises(s_exc.NoSuchProp, t.norm, ('test:str:newp', 'newp'))
        self.raises(s_exc.BadTypeValu, t.norm, ('test:str:tick', '2020', 'a wild argument appears'))

    def test_range(self):
        model = s_datamodel.Model()
        t = model.type('range')

        self.raises(s_exc.BadTypeValu, t.norm, 1)
        self.raises(s_exc.BadTypeValu, t.norm, '1')
        self.raises(s_exc.BadTypeValu, t.norm, (1,))
        self.raises(s_exc.BadTypeValu, t.norm, (1, -1))

        norm, info = t.norm((0, 0))
        self.eq(norm, (0, 0))
        self.eq(info['subs']['min'], 0)
        self.eq(info['subs']['max'], 0)

        self.eq((10, 20), t.norm('10-20')[0])

        norm, info = t.norm((-10, 0xFF))
        self.eq(norm, (-10, 255))
        self.eq(info['subs']['min'], -10)
        self.eq(info['subs']['max'], 255)

        self.eq(t.repr((-10, 0xFF)), ('-10', '255'))

        # Invalid Config
        self.raises(s_exc.BadTypeDef, model.type('range').clone, {'type': None})
        self.raises(s_exc.BadTypeDef, model.type('range').clone, {'type': ('inet:ipv4', {})})  # inet is not loaded yet

    async def test_range_filter(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                node = await snap.addNode('test:str', 'a', {'bar': ('test:str', 'a'), 'tick': '19990101'})
                node = await snap.addNode('test:str', 'b', {'.seen': ('20100101', '20110101'), 'tick': '20151207'})
                node = await snap.addNode('test:str', 'm', {'bar': ('test:str', 'm'), 'tick': '20200101'})
                node = await snap.addNode('test:guid', 'C' * 32)
                node = await snap.addNode('test:guid', 'F' * 32)
                node = await alist(core.eval('[edge:refs=((test:comp, (2048, horton)), (test:comp, (4096, whoville)))]'))
                node = await alist(core.eval('[edge:refs=((test:comp, (9001, "A mean one")), (test:comp, (40000, greeneggs)))]'))
                node = await alist(core.eval('[edge:refs=((test:int, 16), (test:comp, (9999, greenham)))]'))

            nodes = await alist(core.eval('test:str=a +:tick*range=(20000101, 20101201)'))
            self.eq(0, len(nodes))
            nodes = await alist(core.eval('test:str +:tick*range=(19701125, 20151212)'))
            self.eq({node.ndef[1] for node in nodes}, {'a', 'b'})
            nodes = await alist(core.eval('test:comp +:haha*range=(grinch, meanone)'))
            self.eq({node.ndef[1] for node in nodes}, {(2048, 'horton')})
            nodes = await alist(core.eval('test:str +:bar*range=((test:str, c), (test:str, q))'))
            self.eq({node.ndef[1] for node in nodes}, {'m'})
            nodes = await alist(core.eval('test:comp +test:comp*range=((1024, grinch), (4096, zemeanone))'))
            self.eq({node.ndef[1] for node in nodes}, {(2048, 'horton'), (4096, 'whoville')})
            guid0 = 'B' * 32
            guid1 = 'D' * 32
            nodes = await alist(core.eval(f'test:guid +test:guid*range=({guid0}, {guid1})'))
            self.eq({node.ndef[1] for node in nodes}, {'c' * 32})
            nodes = await alist(core.eval('test:int -> test:comp:hehe +test:comp*range=((1000, grinch), (4000, whoville))'))
            self.eq({node.ndef[1] for node in nodes}, {(2048, 'horton')})
            nodes = await alist(core.eval('edge:refs +:n1*range=((test:comp, (1000, green)), (test:comp, (3000, ham)))'))
            self.eq({node.ndef[1] for node in nodes},
                    {(('test:comp', (2048, 'horton')), ('test:comp', (4096, 'whoville')))})

            # The following tests show range working against a string
            nodes = await alist(core.eval('test:str*range=(b, m)'))
            self.len(2, nodes)
            nodes = await alist(core.eval('test:str +test:str*range=(b, m)'))
            self.len(2, nodes)

            # Range against a integer
            async with await core.snap() as snap:
                node = await snap.addNode('test:int', -1)
                node = await snap.addNode('test:int', 0)
                node = await snap.addNode('test:int', 1)
                node = await snap.addNode('test:int', 2)
            nodes = await alist(core.eval('test:int*range=(0, 2)'))
            self.len(3, nodes)
            nodes = await alist(core.eval('test:int +test:int*range=(0, 2)'))
            self.len(3, nodes)

            nodes = await alist(core.eval('test:int*range=(-1, -1)'))
            self.len(1, nodes)
            nodes = await alist(core.eval('test:int +test:int*range=(-1, -1)'))
            self.len(1, nodes)

            # sad path
            await self.agenraises(s_exc.BadCmprValu, core.eval('test:comp +:hehe*range=(0.0.0.0, 1.1.1.1, 6.6.6.6)'))
            await self.agenraises(s_exc.BadCmprValu, core.eval('test:comp +:haha*range=(somestring,) '))
            await self.agenraises(s_exc.BadCmprValu, core.eval('test:str +:bar*range=Foobar'))
            await self.agenraises(s_exc.BadCmprValu, core.eval('test:int +test:int*range=3456'))

    def test_str(self):

        model = s_datamodel.Model()

        lowr = model.type('str').clone({'lower': True})
        self.eq('foo', lowr.norm('FOO')[0])

        self.eq(True, lowr.cmpr('xxherexx', '~=', 'here'))
        self.eq(False, lowr.cmpr('xxherexx', '~=', '^here'))

        self.eq(True, lowr.cmpr('foo', '!=', 'bar'))
        self.eq(False, lowr.cmpr('foo', '!=', 'FOO'))

        self.eq(True, lowr.cmpr('foobar', '^=', 'FOO'))
        self.eq(False, lowr.cmpr('foubar', '^=', 'FOO'))

        regx = model.type('str').clone({'regex': '^[a-f][0-9]+$'})
        self.eq('a333', regx.norm('a333')[0])
        self.raises(s_exc.BadTypeValu, regx.norm, 'A333')

        regl = model.type('str').clone({'regex': '^[a-f][0-9]+$', 'lower': True})
        self.eq('a333', regl.norm('a333')[0])
        self.eq('a333', regl.norm('A333')[0])

        byts = s_common.uhex('e2889e')

        # The real world is a harsh place.
        strp = model.type('str').clone({'strip': True})
        self.eq('foo', strp.norm('  foo \t')[0])

        onespace = model.type('str').clone({'onespace': True})
        self.eq('foo', onespace.norm('  foo\t')[0])
        self.eq('hehe haha', onespace.norm('hehe    haha')[0])

        enums = model.type('str').clone({'enums': 'hehe,haha,zork'})
        self.eq('hehe', enums.norm('hehe')[0])
        self.eq('haha', enums.norm('haha')[0])
        self.eq('zork', enums.norm('zork')[0])
        self.raises(s_exc.BadTypeValu, enums.norm, 'zing')

        strsubs = model.type('str').clone({'regex': r'(?P<first>[ab]+)(?P<last>[zx]+)'})
        norm, info = strsubs.norm('aabbzxxxxxz')
        self.eq(info.get('subs'), {'first': 'aabb', 'last': 'zxxxxxz'})

    def test_syntag(self):

        model = s_datamodel.Model()
        tagtype = model.type('syn:tag')

        self.eq('foo.bar', tagtype.norm(('FOO', ' BAR'))[0])
        self.eq('foo.st_lucia', tagtype.norm(('FOO', 'st.lucia'))[0])

        self.eq('foo.bar', tagtype.norm('FOO.BAR')[0])
        self.eq('foo.bar', tagtype.norm('#foo.bar')[0])
        self.eq('foo.bar', tagtype.norm('foo   .   bar')[0])

        tag, info = tagtype.norm('foo')
        subs = info.get('subs')
        self.none(subs.get('up'))
        self.eq('foo', subs.get('base'))
        self.eq(0, subs.get('depth'))

        tag, info = tagtype.norm('foo.bar')
        subs = info.get('subs')
        self.eq('foo', subs.get('up'))

        self.eq('r_y', tagtype.norm('@#R)(Y')[0])
        self.eq('foo.bar', tagtype.norm('foo\udcfe.bar')[0])
        self.raises(s_exc.BadTypeValu, tagtype.norm, 'foo.')
        self.raises(s_exc.BadTypeValu, tagtype.norm, 'foo..bar')
        self.raises(s_exc.BadTypeValu, tagtype.norm, '.')
        self.raises(s_exc.BadTypeValu, tagtype.norm, '')
        # Tags including non-english unicode letters are okay
        self.eq('icon.ॐ', tagtype.norm('ICON.ॐ')[0])
        # homoglyphs are also possible
        self.eq('is.ｂob.evil', tagtype.norm('is.\uff42ob.evil')[0])

    async def test_time(self):

        model = s_datamodel.Model()
        ttime = model.types.get('time')

        with self.raises(s_exc.BadTypeValu):
            ttime.norm('0000-00-00')

        self.gt(s_common.now(), ttime.norm('-1hour')[0])

        tminmax = ttime.clone({'min': True, 'max': True})
        # Merge testing with tminmax
        now = s_common.now()
        self.eq(now + 1, tminmax.merge(now, now + 1))
        self.eq(now, tminmax.merge(now + 1, now))

        async with self.getTestCore() as core:

            t = core.model.type('test:time')

            # explicitly test our "future/ongoing" value...
            future = 0x7fffffffffffffff
            self.eq(t.norm('?')[0], future)
            self.eq(t.norm(future)[0], future)
            self.eq(t.repr(future), '?')

            # Explicitly test our max time vs. future marker
            maxtime = 253402300799999  # 9999/12/31 23:59:59.999
            self.eq(t.norm(maxtime)[0], maxtime)
            self.eq(t.repr(maxtime), '9999/12/31 23:59:59.999')
            self.eq(t.norm('9999/12/31 23:59:59.999')[0], maxtime)
            self.raises(s_exc.BadTypeValu, t.norm, maxtime + 1)

            tick = t.norm('2014')[0]
            self.eq(t.repr(tick), '2014/01/01 00:00:00.000')

            tock = t.norm('2015')[0]

            self.raises(s_exc.BadCmprValu,
                        t.cmpr, '2015', 'range=', tick)

            async with await core.snap() as snap:
                node = await snap.addNode('test:str', 'a', {'tick': '2014'})
                node = await snap.addNode('test:str', 'b', {'tick': '2015'})
                node = await snap.addNode('test:str', 'c', {'tick': '2016'})
                node = await snap.addNode('test:str', 'd', {'tick': 'now'})

            nodes = await core.nodes('test:str:tick=2014')
            self.eq({node.ndef[1] for node in nodes}, {'a'})

            nodes = await core.nodes('test:str:tick*range=(2014, 2015)')
            self.eq({node.ndef[1] for node in nodes}, {'a', 'b'})

            nodes = await core.nodes('test:str:tick=201401*')
            self.eq({node.ndef[1] for node in nodes}, {'a'})

            nodes = await core.nodes('test:str:tick*range=("-4200 days", now)')
            self.eq({node.ndef[1] for node in nodes}, {'a', 'b', 'c', 'd'})

            opts = {'vars': {'tick': tick, 'tock': tock}}
            nodes = await core.nodes('test:str:tick*range=($tick, $tock)', opts=opts)
            self.eq({node.ndef[1] for node in nodes}, {'a', 'b'})

            nodes = await core.nodes('test:str:tick*range=(20131231, "+2 days")')
            self.eq({node.ndef[1] for node in nodes}, {'a'})

            nodes = await core.nodes('test:str:tick*range=("-1 day", "+1 day")')
            self.eq({node.ndef[1] for node in nodes}, {'d'})

            nodes = await core.nodes('test:str:tick*range=("-1 days", now)')
            self.eq({node.ndef[1] for node in nodes}, {'d'})

            # Equivalent lift
            nodes = await core.nodes('test:str:tick*range=(now, "-1 days")')
            self.eq({node.ndef[1] for node in nodes}, {'d'})
            # This is equivalent of the previous lift

            self.eq({node.ndef[1] for node in nodes}, {'d'})

            self.true(t.cmpr('2015', '>=', '20140202'))
            self.true(t.cmpr('2015', '>=', '2015'))
            self.true(t.cmpr('2015', '>', '20140202'))
            self.false(t.cmpr('2015', '>', '2015'))

            self.true(t.cmpr('20150202', '<=', '2016'))
            self.true(t.cmpr('20150202', '<=', '2016'))
            self.true(t.cmpr('20150202', '<', '2016'))
            self.false(t.cmpr('2015', '<', '2015'))

            await self.agenlen(1, core.eval('test:str +:tick=2015'))

            await self.agenlen(1, core.eval('test:str +:tick*range=($test, "+- 2day")',
                                            opts={'vars': {'test': '2015'}}))

            await self.agenlen(1, core.eval('test:str +:tick*range=(now, "-+ 1day")'))

            await self.agenlen(1, core.eval('test:str +:tick*range=(2015, "+1 day")'))
            await self.agenlen(1, core.eval('test:str +:tick*range=(20150102, "-3 day")'))
            await self.agenlen(0, core.eval('test:str +:tick*range=(20150201, "+1 day")'))
            await self.agenlen(1, core.eval('test:str +:tick*range=(20150102, "+- 2day")'))
            await self.agenlen(2, core.eval('test:str +:tick*range=(2015, 2016)'))
            await self.agenlen(0, core.eval('test:str +:tick*range=(2016, 2015)'))

            await self.agenlen(2, core.eval('test:str:tick*range=(2015, 2016)'))
            await self.agenlen(0, core.eval('test:str:tick*range=(2016, 2015)'))

            await self.agenlen(1, core.eval('test:str:tick*range=(2015, "+1 day")'))
            await self.agenlen(4, core.eval('test:str:tick*range=(2014, "now")'))
            await self.agenlen(0, core.eval('test:str:tick*range=(20150201, "+1 day")'))
            await self.agenlen(1, core.eval('test:str:tick*range=(now, "+-1 day")'))
            await self.agenlen(0, core.eval('test:str:tick*range=(now, "+1 day")'))

            # Sad path for *range=
            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:str:tick*range=("+- 1day", "now")')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:str:tick*range=("-+ 1day", "now")')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('[test:guid="*" :tick="+-1 day"]')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:str:tick*range=(2015)')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:str:tick*range=$tick', opts={'vars': {'tick': tick}})

            await self.agenraises(s_exc.BadCmprValu, core.eval('test:str +:tick*range=(2015)'))
            await self.agenraises(s_exc.BadCmprValu, core.eval('test:str +:tick*range=(2015, 2016, 2017)'))
            await self.agenraises(s_exc.BadTypeValu, core.eval('test:str +:tick*range=("?", "+1 day")'))
            await self.agenraises(s_exc.BadTypeValu, core.eval('test:str +:tick*range=(2000, "?+1 day")'))
            await self.agenraises(s_exc.BadTypeValu, core.eval('test:str:tick*range=("?", "+1 day")'))
            await self.agenraises(s_exc.BadTypeValu, core.eval('test:str:tick*range=(2000, "?+1 day")'))

            async with await core.snap() as snap:
                node = await snap.addNode('test:str', 't1', {'tick': '2018/12/02 23:59:59.000'})
                node = await snap.addNode('test:str', 't2', {'tick': '2018/12/03'})
                node = await snap.addNode('test:str', 't3', {'tick': '2018/12/03 00:00:01.000'})

            await self.agenlen(0, core.eval('test:str:tick*range=(2018/12/01, "+24 hours")'))
            await self.agenlen(2, core.eval('test:str:tick*range=(2018/12/01, "+48 hours")'))

            await self.agenlen(0, core.eval('test:str:tick*range=(2018/12/02, "+23 hours")'))
            await self.agenlen(1, core.eval('test:str:tick*range=(2018/12/02, "+86399 seconds")'))
            await self.agenlen(2, core.eval('test:str:tick*range=(2018/12/02, "+24 hours")'))
            await self.agenlen(3, core.eval('test:str:tick*range=(2018/12/02, "+86401 seconds")'))
            await self.agenlen(3, core.eval('test:str:tick*range=(2018/12/02, "+25 hours")'))

            self.len(0, await core.nodes('test:guid | delnode --force'))
            await core.nodes('[ test:guid=* :tick=20211031020202 ]')

            # test * suffix syntax
            self.len(1, await core.nodes('test:guid:tick=2021*'))
            self.len(1, await core.nodes('test:guid:tick=202110*'))
            self.len(1, await core.nodes('test:guid:tick=20211031*'))
            self.len(1, await core.nodes('test:guid:tick=2021103102*'))
            self.len(1, await core.nodes('test:guid:tick=202110310202*'))
            self.len(1, await core.nodes('test:guid:tick=20211031020202*'))

            self.len(0, await core.nodes('test:guid:tick=2022*'))
            self.len(0, await core.nodes('test:guid:tick=202210*'))
            self.len(0, await core.nodes('test:guid:tick=20221031*'))
            self.len(0, await core.nodes('test:guid:tick=2022103102*'))
            self.len(0, await core.nodes('test:guid:tick=202210310202*'))
            self.len(0, await core.nodes('test:guid:tick=20221031020202*'))

            self.len(1, await core.nodes('test:guid +:tick=2021*'))
            self.len(1, await core.nodes('test:guid +:tick=202110*'))
            self.len(1, await core.nodes('test:guid +:tick=20211031*'))
            self.len(1, await core.nodes('test:guid +:tick=2021103102*'))
            self.len(1, await core.nodes('test:guid +:tick=202110310202*'))
            self.len(1, await core.nodes('test:guid +:tick=20211031020202*'))

            self.len(0, await core.nodes('test:guid -:tick=2021*'))
            self.len(0, await core.nodes('test:guid -:tick=202110*'))
            self.len(0, await core.nodes('test:guid -:tick=20211031*'))
            self.len(0, await core.nodes('test:guid -:tick=2021103102*'))
            self.len(0, await core.nodes('test:guid -:tick=202110310202*'))
            self.len(0, await core.nodes('test:guid -:tick=20211031020202*'))

            # test <= time * suffix syntax
            self.len(1, await core.nodes('test:guid:tick<=2021*'))
            self.len(1, await core.nodes('test:guid:tick<=202110*'))
            self.len(1, await core.nodes('test:guid:tick<=20211031*'))
            self.len(1, await core.nodes('test:guid:tick<=2021103102*'))
            self.len(1, await core.nodes('test:guid:tick<=202110310202*'))
            self.len(1, await core.nodes('test:guid:tick<=20211031020202*'))

            self.len(0, await core.nodes('test:guid:tick<=2020*'))
            self.len(0, await core.nodes('test:guid:tick<=202010*'))
            self.len(0, await core.nodes('test:guid:tick<=20201031*'))
            self.len(0, await core.nodes('test:guid:tick<=2020103102*'))
            self.len(0, await core.nodes('test:guid:tick<=202010310202*'))
            self.len(0, await core.nodes('test:guid:tick<=20201031020202*'))

            self.len(1, await core.nodes('test:guid +:tick<=2021*'))
            self.len(1, await core.nodes('test:guid +:tick<=202110*'))
            self.len(1, await core.nodes('test:guid +:tick<=20211031*'))
            self.len(1, await core.nodes('test:guid +:tick<=2021103102*'))
            self.len(1, await core.nodes('test:guid +:tick<=202110310202*'))
            self.len(1, await core.nodes('test:guid +:tick<=20211031020202*'))

            self.len(0, await core.nodes('test:guid -:tick<=2021*'))
            self.len(0, await core.nodes('test:guid -:tick<=202110*'))
            self.len(0, await core.nodes('test:guid -:tick<=20211031*'))
            self.len(0, await core.nodes('test:guid -:tick<=2021103102*'))
            self.len(0, await core.nodes('test:guid -:tick<=202110310202*'))
            self.len(0, await core.nodes('test:guid -:tick<=20211031020202*'))

            # test <= time * suffix syntax
            self.len(1, await core.nodes('test:guid:tick<2021*'))
            self.len(1, await core.nodes('test:guid:tick<202110*'))
            self.len(1, await core.nodes('test:guid:tick<20211031*'))
            self.len(1, await core.nodes('test:guid:tick<2021103102*'))
            self.len(1, await core.nodes('test:guid:tick<202110310202*'))
            self.len(1, await core.nodes('test:guid:tick<20211031020202*'))

            self.len(0, await core.nodes('test:guid:tick<2020*'))
            self.len(0, await core.nodes('test:guid:tick<202010*'))
            self.len(0, await core.nodes('test:guid:tick<20201031*'))
            self.len(0, await core.nodes('test:guid:tick<2020103102*'))
            self.len(0, await core.nodes('test:guid:tick<202010310202*'))
            self.len(0, await core.nodes('test:guid:tick<20201031020202*'))

            self.len(1, await core.nodes('test:guid +:tick<2021*'))
            self.len(1, await core.nodes('test:guid +:tick<202110*'))
            self.len(1, await core.nodes('test:guid +:tick<20211031*'))
            self.len(1, await core.nodes('test:guid +:tick<2021103102*'))
            self.len(1, await core.nodes('test:guid +:tick<202110310202*'))
            self.len(1, await core.nodes('test:guid +:tick<20211031020202*'))

            self.len(0, await core.nodes('test:guid -:tick<2021*'))
            self.len(0, await core.nodes('test:guid -:tick<202110*'))
            self.len(0, await core.nodes('test:guid -:tick<20211031*'))
            self.len(0, await core.nodes('test:guid -:tick<2021103102*'))
            self.len(0, await core.nodes('test:guid -:tick<202110310202*'))
            self.len(0, await core.nodes('test:guid -:tick<20211031020202*'))

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:guid:tick=202*')

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('test:guid:tick=202110310202021*')

        async with self.getTestCore() as core:

            async with await core.snap() as snap:
                node = await snap.addNode('test:str', 'a', {'tick': '2014'})
                node = await snap.addNode('test:int', node.get('tick') + 1)
                node = await snap.addNode('test:str', 'b', {'tick': '2015'})
                node = await snap.addNode('test:int', node.get('tick') + 1)
                node = await snap.addNode('test:str', 'c', {'tick': '2016'})
                node = await snap.addNode('test:int', node.get('tick') + 1)
                node = await snap.addNode('test:str', 'd', {'tick': 'now'})
                node = await snap.addNode('test:int', node.get('tick') + 1)

            q = 'test:int $end=$node.value() test:str:tick*range=(2015, $end) -test:int'
            nodes = await alist(core.eval(q))
            self.len(6, nodes)
            self.eq({node.ndef[1] for node in nodes}, {'b', 'c', 'd'})

    async def test_edges(self):
        model = s_datamodel.Model()
        e = model.type('edge')
        t = model.type('timeedge')

        self.eq(0, e.getCompOffs('n1'))
        self.eq(1, e.getCompOffs('n2'))
        self.none(e.getCompOffs('newp'))

        self.eq(0, t.getCompOffs('n1'))
        self.eq(1, t.getCompOffs('n2'))
        self.eq(2, t.getCompOffs('time'))
        self.none(t.getCompOffs('newp'))

        # Sad path testing
        self.raises(s_exc.BadTypeValu, e.norm, ('newp',))
        self.raises(s_exc.BadTypeValu, t.norm, ('newp',))

        # Repr testing with the test model
        async with self.getTestCore() as core:
            e = core.model.type('edge')
            t = core.model.type('timeedge')

            norm, _ = e.norm((('test:str', '1234'), ('test:int', '1234')))
            self.eq(norm, (('test:str', '1234'), ('test:int', 1234)))

            norm, _ = t.norm((('test:str', '1234'), ('test:int', '1234'), '2001'))
            self.eq(norm, (('test:str', '1234'), ('test:int', 1234), 978307200000))

            rval = e.repr((('test:str', '1234'), ('test:str', 'hehe')))
            self.eq(rval, (('test:str', '1234'), ('test:str', 'hehe')))

            rval = e.repr((('test:int', 1234), ('test:str', 'hehe')))
            self.eq(rval, (('test:int', '1234'), ('test:str', 'hehe')))

            rval = e.repr((('test:str', 'hehe'), ('test:int', 1234)))
            self.eq(rval, (('test:str', 'hehe'), ('test:int', '1234')))

            rval = e.repr((('test:int', 4321), ('test:int', 1234)))
            self.eq(rval, (('test:int', '4321'), ('test:int', '1234')))

            rval = e.repr((('test:int', 4321), ('test:comp', (1234, 'hehe'))))
            self.eq(rval, (('test:int', '4321'), ('test:comp', ('1234', 'hehe'))))

            tv = 5356800000
            tr = '1970/03/04 00:00:00.000'

            rval = t.repr((('test:str', '1234'), ('test:str', 'hehe'), tv))
            self.eq(rval, (('test:str', '1234'), ('test:str', 'hehe'), tr))

            rval = t.repr((('test:int', 1234), ('test:str', 'hehe'), tv))
            self.eq(rval, (('test:int', '1234'), ('test:str', 'hehe'), tr))

            rval = t.repr((('test:str', 'hehe'), ('test:int', 1234), tv))
            self.eq(rval, (('test:str', 'hehe'), ('test:int', '1234'), tr))

    async def test_types_long_indx(self):

        aaaa = 'A' * 200
        url = f'http://vertex.link/visi?fuzz0={aaaa}&fuzz1={aaaa}'

        async with self.getTestCore() as core:
            opts = {'vars': {'url': url}}
            self.len(1, await core.nodes('[ it:exec:url="*" :url=$url ]', opts=opts))
            self.len(1, await core.nodes('it:exec:url:url=$url', opts=opts))

    async def test_types_array(self):

        mdef = {
            'types': (
                ('test:array', ('array', {'type': 'inet:ipv4'}), {}),
                ('test:arraycomp', ('comp', {'fields': (('ipv4s', 'test:array'), ('int', 'test:int'))}), {}),
                ('test:witharray', ('guid', {}), {}),
            ),
            'forms': (
                ('test:array', {}, (
                )),
                ('test:arraycomp', {}, (
                    ('ipv4s', ('test:array', {}), {}),
                    ('int', ('test:int', {}), {}),
                )),
                ('test:witharray', {}, (
                    ('fqdns', ('array', {'type': 'inet:fqdn', 'uniq': True, 'sorted': True, 'split': ','}), {}),
                )),
            ),
        }
        async with self.getTestCore() as core:

            core.model.addDataModels([('asdf', mdef)])

            with self.raises(s_exc.BadTypeDef):
                await core.addFormProp('test:int', '_hehe', ('array', {'type': 'array'}), {})

            with self.raises(s_exc.BadTypeDef):
                await core.addFormProp('test:int', '_hehe', ('array', {'type': 'newp'}), {})

            nodes = await core.nodes('[ test:array=(1.2.3.4, 5.6.7.8) ]')
            self.len(1, nodes)

            # create a long array (fails pre-020)
            arr = ','.join([str(i) for i in range(300)])
            nodes = await core.nodes(f'[ test:array=({arr}) ]')
            self.len(1, nodes)

            nodes = await core.nodes('test:array*[=1.2.3.4]')
            self.len(1, nodes)

            nodes = await core.nodes('test:array*[=1.2.3.4] | delnode')
            nodes = await core.nodes('test:array*[=1.2.3.4]')
            self.len(0, nodes)

            with self.raises(s_exc.BadTypeValu):
                await core.nodes('[ test:arraycomp=("1.2.3.4, 5.6.7.8", 10) ]')

            nodes = await core.nodes('[ test:arraycomp=((1.2.3.4, 5.6.7.8), 10) ]')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('test:arraycomp', ((0x01020304, 0x05060708), 10)))
            self.eq(nodes[0].get('int'), 10)
            self.eq(nodes[0].get('ipv4s'), (0x01020304, 0x05060708))

            # make sure "adds" got added
            nodes = await core.nodes('inet:ipv4=1.2.3.4 inet:ipv4=5.6.7.8')
            self.len(2, nodes)

            nodes = await core.nodes('[ test:witharray="*" :fqdns="woot.com, VERTEX.LINK, vertex.link" ]')
            self.len(1, nodes)

            self.eq(nodes[0].get('fqdns'), ('vertex.link', 'woot.com'))
            self.sorteq(('vertex.link', 'woot.com'), nodes[0].repr('fqdns').split(','))

            nodes = await core.nodes('test:witharray:fqdns=(vertex.link, WOOT.COM)')
            self.len(1, nodes)

            nodes = await core.nodes('test:witharray:fqdns*[=vertex.link]')
            self.len(1, nodes)

            nodes = await core.nodes('test:witharray [ :fqdns=(hehe.com, haha.com) ]')
            self.len(1, nodes)

            nodes = await core.nodes('inet:fqdn=hehe.com inet:fqdn=haha.com')
            self.len(2, nodes)

            # make sure the multi-array entries got deleted
            nodes = await core.nodes('test:witharray:fqdns*[=vertex.link]')
            self.len(0, nodes)

            nodes = await core.nodes('test:witharray:fqdns*[=hehe.com]')
            self.len(1, nodes)

            nodes = await core.nodes('test:witharray:fqdns*[~=ehe]')
            self.len(1, nodes)



import json

import logging

import synapse.common as s_common

class JsonFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record: logging.LogRecord):

        record.message = record.getMessage()
        mesg = self.formatMessage(record)
        ret = {
            'message': mesg,
            'logger': {
                'name': record.name,
                'process': record.processName,
                'filename': record.filename,
                'func': record.funcName,
            },
            'level': record.levelname,
            'time': self.formatTime(record, self.datefmt),
        }

        if record.exc_info:
            name, info = s_common.err(record.exc_info[1], fulltb=True)
            # This is the actual exception name. The ename key is the function name.
            info['errname'] = name
            ret['err'] = info

        # stuffing our extra into a single dictionary avoids a loop
        # over record.__dict__ extracting fields which are not known
        # attributes for each log record.
        extras = record.__dict__.get('synapse')
        if extras:
            ret.update({k: v for k, v in extras.items() if k not in ret})

        return json.dumps(ret, default=str)



import synapse.common as s_common

import synapse.tests.utils as s_t_utils
import synapse.tools.guid as s_guid

class TestGuid(s_t_utils.SynTest):

    def test_tools_guid(self):
        argv = []
        outp = self.getTestOutp()
        s_guid.main(argv, outp=outp)
        self.true(s_common.isguid(str(outp)))



import enum
import time
import heapq
import asyncio
import logging
import calendar
import datetime
import functools
import itertools
import collections
from datetime import timezone as tz
from collections.abc import Iterable, Mapping

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.base as s_base
import synapse.lib.coro as s_coro
import synapse.lib.config as s_config
import synapse.lib.provenance as s_provenance

# Agenda: manages running one-shot and periodic tasks in the future ("appointments")

logger = logging.getLogger(__name__)

reqValidCdef = s_config.getJsValidator({
    'type': 'object',
    'properties': {
        'storm': {'type': 'string'},
        'creator': {'type': 'string', 'pattern': s_config.re_iden},
        'iden': {'type': 'string', 'pattern': s_config.re_iden},
        'view': {'type': 'string', 'pattern': s_config.re_iden},
        'name': {'type': 'string'},
        'doc': {'type': 'string'},
        'incunit': {
            'oneOf': [
                {'type': 'null'},
                {'enum': ['year', 'month', 'dayofmonth', 'dayofweek', 'day', 'hour', 'minute']}
            ]
        },
        'incvals': {
            'type': ['array', 'number', 'null'],
            'items': {'type': 'number'}
        },
        'reqs': {
            'oneOf': [
                {
                    '$ref': '#/definitions/req',
                },
                {
                    'type': ['array'],
                    'items': {'$ref': '#/definitions/req'},
                },
            ]
        },
    },
    'additionalProperties': False,
    'required': ['creator', 'storm'],
    'dependencices': {
        'incvals': ['incunit'],
        'incunit': ['incvals'],
    },
    'definitions': {
        'req': {
            'type': 'object',
            'properties': {
                'minute': {'oneOf': [{'type': 'number'}, {'type': 'array', 'items': {'type': 'number'}}]},
                'hour': {'oneOf': [{'type': 'number'}, {'type': 'array', 'items': {'type': 'number'}}]},
                'dayofmonth': {'oneOf': [{'type': 'number'}, {'type': 'array', 'items': {'type': 'number'}}]},
                'month': {'oneOf': [{'type': 'number'}, {'type': 'array', 'items': {'type': 'number'}}]},
                'year': {'oneOf': [{'type': 'number'}, {'type': 'array', 'items': {'type': 'number'}}]},
            }
        }
    }
})

def _dayofmonth(hardday, month, year):
    '''
    Returns a valid day of the month given the desired value.

    Negative values are interpreted as offset backwards from the last day of the month, with -1 representing the
    last day of the month.  Out-of-range values are clamped to the first or last day of the month.
    '''
    newday = hardday
    daysinmonth = calendar.monthrange(year, month)[1]
    if newday < 0:
        newday = daysinmonth + hardday + 1
    newday = max(1, min(newday, daysinmonth))
    return newday

class TimeUnit(enum.IntEnum):
    '''
    Unit of time that recurring and required parts of appointments are made of
    '''
    YEAR = enum.auto()
    MONTH = enum.auto()
    DAYOFMONTH = enum.auto()  # e.g. every 3rd of the month
    DAYOFWEEK = enum.auto()   # e.g. every Tuesday (Mon=0, Sun=6)
    DAY = enum.auto()         # every day
    HOUR = enum.auto()
    MINUTE = enum.auto()
    NOW = enum.auto()

    @classmethod
    def fromString(cls, s):
        return cls.__members__[s.upper()]

# Next largest unit for each unit
_NextUnitMap = {
    TimeUnit.YEAR: None,
    TimeUnit.MONTH: TimeUnit.YEAR,
    TimeUnit.HOUR: TimeUnit.DAY,
    TimeUnit.MINUTE: TimeUnit.HOUR,
    TimeUnit.DAYOFMONTH: TimeUnit.MONTH,
    TimeUnit.DAYOFWEEK: TimeUnit.DAYOFWEEK,
}

# Unit equivalence to datetime arguments
_TimeunitToDatetime = {
    TimeUnit.YEAR: 'year',
    TimeUnit.MONTH: 'month',
    TimeUnit.DAYOFMONTH: 'day',
    TimeUnit.DAYOFWEEK: 'day',
    TimeUnit.DAY: 'day',
    TimeUnit.HOUR: 'hour',
    TimeUnit.MINUTE: 'minute',
}

# The valid ranges for required and recurring
_UnitBounds = {
    TimeUnit.YEAR: ((2000, 2999), (1, 5)),
    TimeUnit.MONTH: ((1, 12), (1, 60)),
    TimeUnit.DAYOFMONTH: ((-31, 31), (-31, 31)),
    TimeUnit.DAYOFWEEK: ((0, 6), (0, 6)),
    TimeUnit.DAY: ((0, 0), (1, 365 * 5)),
    TimeUnit.HOUR: ((0, 23), (1, 1000)),
    TimeUnit.MINUTE: ((0, 59), (1, 60 * 60 * 24 * 30))
}

class ApptRec:
    '''
    Represents a single element of a single combination of an appointment
    '''
    def __init__(self, reqdict, incunit=None, incval=1):
        if incunit is not None:
            incunit = TimeUnit(incunit)
        self.incunit = incunit
        self.incval = incval if incunit is not None else None

        if not reqdict and incunit is None:
            raise s_exc.BadTime(mesg='reqdict must be nonempty or incunit must be non-None')

        if TimeUnit.DAY in reqdict:
            raise s_exc.BadTime(mesg='Must not specify day as requirement')

        if TimeUnit.DAYOFMONTH in reqdict and TimeUnit.DAYOFWEEK in reqdict:
            raise s_exc.BadTime(mesg='Both day of month and day of week must not both be requirements')

        if TimeUnit.DAYOFWEEK in reqdict and incunit is not None:
            raise s_exc.BadTime(mesg='Day of week requirement not supported with a recurrence')

        if incunit == TimeUnit.DAYOFMONTH:
            raise s_exc.BadTime(mesg='Day of month not a valid incunit')

        if incunit is not None:
            boundmin, boundmax = _UnitBounds[incunit][1]
            if not boundmin <= incval <= boundmax:
                raise s_exc.BadTime(mesg='Out of bounds incval')

        reqdict = {TimeUnit(k): v for k, v in reqdict.items()}
        self.reqdict = reqdict

        for reqkey, reqval in reqdict.items():
            boundmin, boundmax = _UnitBounds[reqkey][0]
            if not boundmin <= reqval <= boundmax:
                raise s_exc.BadTime(mesg='Out of bounds reqdict value')

            if incunit is not None and reqkey <= incunit:
                # We could, actually support this, but most of these combinations are nonsensical (e.g. run every 5
                # minutes in 2018 only?)
                raise s_exc.BadTime(mesg='Must not have fixed unit equal to or greater than recurrence unit')

        self.reqdict = {}
        # Put keys in size order, with dayof... added last, as nexttime processes in that order
        for key in _NextUnitMap:
            if key in reqdict:
                self.reqdict[key] = reqdict[key]

    def __repr__(self):
        return repr(self.pack())

    def pack(self):
        '''
        Make ApptRec json/msgpack-friendly
        '''
        reqdictf = {k.name.lower(): v for (k, v) in self.reqdict.items()}
        incunitf = None if self.incunit is None else self.incunit.name.lower()
        return (reqdictf, incunitf, self.incval)

    @classmethod
    def unpack(cls, val):
        '''
        Convert from json/msgpack-friendly
        '''
        reqdictf, incunitf, incval = val
        reqdict = {TimeUnit[k.upper()]: v for (k, v) in reqdictf.items()}
        incunit = None if incunitf is None else TimeUnit[incunitf.upper()]
        return cls(reqdict, incunit, incval)

    def nexttime(self, lastts):
        '''
        Returns next timestamp that meets requirements, incrementing by (self.incunit * incval) if not increasing, or
        0.0 if there are no future matches
        '''
        lastdt = datetime.datetime.fromtimestamp(lastts, tz.utc)
        newvals = {}  # all the new fields that will be changed in datetime of lastts

        # Truncate the seconds part
        newdt = lastdt.replace(second=0)

        # Note: self.reqdict is sorted from largest unit to smallest
        for unit, newval in self.reqdict.items():
            dtkey = _TimeunitToDatetime[unit]

            if unit is TimeUnit.DAYOFWEEK:
                newdt = newdt.replace(**newvals)
                newvals = {}
                newval = newdt.day + (6 + newval - newdt.weekday()) % 7 + 1
                if newval > calendar.monthrange(newdt.year, newdt.month)[1]:
                    newval -= 7

            elif unit is TimeUnit.YEAR:
                # As we change the year, clamp the day of the month to a valid value (only matters on leap day)
                dayval = _dayofmonth(newdt.day, newdt.month, newval)
                newvals['day'] = dayval

            elif unit is TimeUnit.MONTH:
                # As we change the month, clamp the day of the month to a valid value
                newdt = newdt.replace(**newvals)
                newvals = {}
                dayval = _dayofmonth(newdt.day, newval, newdt.year)
                newvals['day'] = dayval

            elif unit is TimeUnit.DAYOFMONTH:
                newdt = newdt.replace(**newvals)
                newvals = {}
                newval = _dayofmonth(newval, newdt.month, newdt.year)

            newvals[dtkey] = newval

        newdt = newdt.replace(**newvals)

        # Then move forward if we have to
        if newdt <= lastdt or \
                self.incunit == TimeUnit.DAYOFWEEK and newdt.weekday() != self.incval:
            if self.incunit is None:
                largest_req = min(self.reqdict.keys())
                tmpunit = _NextUnitMap[largest_req]
                if tmpunit is None:  # required a year and we're already there
                    return 0.0
                # Unless we're going to the next day of week, increment by 1 unit of the next larger unit
                tmpincval = self.reqdict.get(TimeUnit.DAYOFWEEK, 1)
            else:
                tmpunit = self.incunit
                tmpincval = self.incval
            newdt = self._inc(tmpunit, tmpincval, self.reqdict, lastdt, newdt)
            assert newdt > lastdt
        return newdt.timestamp()

    def _inc(self, incunit, incval, reqdict, origdt, dt):
        '''
        Return a datetime incremented by incunit * incval
        '''
        if incunit == TimeUnit.YEAR:
            return dt.replace(year=dt.year + incval)
        if incunit == TimeUnit.MONTH:
            newyear = dt.year
            absmonth = dt.month + incval - 1
            newmonth = absmonth % 12 + 1
            newyear += absmonth // 12
            daysinmonth = calendar.monthrange(newyear, newmonth)[1]
            dayofmonthreq = reqdict.get(TimeUnit.DAYOFMONTH)
            if dayofmonthreq is not None:
                newday = _dayofmonth(dayofmonthreq, newmonth, newyear)
            else:
                newday = min(daysinmonth, dt.day)
            return dt.replace(day=newday, month=newmonth, year=newyear)
        if incunit == TimeUnit.DAY:
            return dt + datetime.timedelta(days=incval)
        if incunit == TimeUnit.DAYOFWEEK:
            # incval in this case means next day of week whose weekday matches incval (0-6)
            days = (6 + incval - dt.weekday()) % 7 + 1
            newdt = dt + datetime.timedelta(days=days)
            assert newdt.weekday() == incval
            return newdt
        if incunit == TimeUnit.HOUR:
            return dt + datetime.timedelta(hours=incval)
        if incunit == TimeUnit.MINUTE:
            return dt + datetime.timedelta(minutes=incval)
        else:
            assert 0, 'Invalid incunit'  # pragma: no cover

class _Appt:
    '''
    A single entry in the Agenda:  a storm query to run in the future, potentially more than once

    Each such entry has a list of ApptRecs.  Each time the appointment is scheduled, the nexttime of the appointment is
    the lowest nexttime of all its ApptRecs.
    '''
    def __init__(self, stor, iden, recur, indx, query, creator, recs, nexttime=None, view=None):
        self.doc = ''
        self.name = ''
        self.stor = stor
        self.iden = iden
        self.recur = recur  # does this appointment repeat
        self.indx = indx  # incremented for each appt added ever.  Used for nexttime tiebreaking for stable ordering
        self.query = query  # query to run
        self.creator = creator  # user iden to run query as
        self.recs = recs  # List[ApptRec]  list of the individual entries to calculate next time from
        self._recidxnexttime = None  # index of rec who is up next
        self.view = view

        if self.recur and not self.recs:
            raise s_exc.BadTime(mesg='A recurrent appointment with no records')

        if nexttime is None and self.recs:
            self.nexttime = self.stor._getNowTick()
            self.updateNexttime(self.nexttime)
            if self.nexttime is None:
                raise s_exc.BadTime(mesg='Appointment is in the past')
        else:
            self.nexttime = nexttime
        self.isrunning = False  # whether it is currently running
        self.startcount = 0  # how many times query has started
        self.errcount = 0  # how many times this appt failed
        self.lasterrs = collections.deque((), maxlen=5)
        self.laststarttime = None
        self.lastfinishtime = None
        self.lastresult = None
        self.enabled = True

    def getStorNode(self, form):
        ndef = (form.name, form.type.norm(self.iden)[0])
        buid = s_common.buid(ndef)

        props = {
            'doc': self.doc,
            'name': self.name,
            'storm': self.query,
        }

        pnorms = {}
        for prop, valu in props.items():
            formprop = form.props.get(prop)
            if formprop is not None and valu is not None:
                pnorms[prop] = formprop.type.norm(valu)[0]

        return (buid, {
            'ndef': ndef,
            'props': pnorms
        })

    def __eq__(self, other):
        ''' For heap logic to sort upcoming events lower '''
        return (self.nexttime, self.indx) == (other.nexttime, other.indx)

    def __lt__(self, other):
        ''' For heap logic '''
        return (self.nexttime, self.indx) < (other.nexttime, other.indx)

    def pack(self):
        return {
            'ver': 1,
            'doc': self.doc,
            'name': self.name,
            'enabled': self.enabled,
            'recur': self.recur,
            'iden': self.iden,
            'view': self.view,
            'indx': self.indx,
            'query': self.query,
            'creator': self.creator,
            'recs': [d.pack() for d in self.recs],
            'nexttime': self.nexttime,
            'startcount': self.startcount,
            'errcount': self.errcount,
            'isrunning': self.isrunning,
            'laststarttime': self.laststarttime,
            'lastfinishtime': self.lastfinishtime,
            'lastresult': self.lastresult,
            'lasterrs': list(self.lasterrs)
        }

    @classmethod
    def unpack(cls, stor, val):
        if val['ver'] != 1:
            raise s_exc.BadStorageVersion(mesg=f"Found version {val['ver']}")  # pragma: no cover
        recs = [ApptRec.unpack(tupl) for tupl in val['recs']]
        appt = cls(stor, val['iden'], val['recur'], val['indx'], val['query'], val['creator'], recs, nexttime=val['nexttime'], view=val.get('view'))
        appt.doc = val.get('doc', '')
        appt.name = val.get('name', '')
        appt.laststarttime = val['laststarttime']
        appt.lastfinishtime = val['lastfinishtime']
        appt.lastresult = val['lastresult']
        appt.enabled = val['enabled']

        return appt

    def updateNexttime(self, now):
        '''
        Find the next time this appointment should be scheduled.

        Delete any nonrecurring record that just happened.
        '''
        if self._recidxnexttime is not None and not self.recur:
            del self.recs[self._recidxnexttime]

        while self.recs and self.nexttime <= now:

            lowtime = 999999999999.9

            # Find the lowest next time of all of our recs (backwards, so we can delete)
            for i in range(len(self.recs) - 1, -1, -1):
                rec = self.recs[i]
                nexttime = rec.nexttime(self.nexttime)
                if nexttime == 0.0:
                    # We blew by and missed a fixed-year appointment, either due to clock shenanigans, this query going
                    # really long, or the initial requirement being in the past
                    logger.warning(f'Missed an appointment: {rec}')
                    del self.recs[i]
                    continue
                if nexttime < lowtime:
                    lowtime = nexttime
                    lowidx = i

            if not self.recs:
                break

            self._recidxnexttime = lowidx
            self.nexttime = lowtime

        if not self.recs:
            self._recidxnexttime = None
            self.nexttime = None
            return

    async def setDoc(self, text, nexs=False):
        '''
        Set the doc field of an appointment.
        '''
        self.doc = text
        await self._save(nexs=nexs)

    async def setName(self, text, nexs=False):
        self.name = text
        await self._save(nexs=nexs)

    async def _save(self, nexs=False):
        await self.stor._storeAppt(self, nexs=nexs)

class Agenda(s_base.Base):
    '''
    Organize and execute all the scheduled storm queries in a cortex.
    '''

    async def __anit__(self, core):

        await s_base.Base.__anit__(self)

        self.core = core
        self.apptheap = []  # Stores the appointments in a heap such that the first element is the next appt to run
        self.appts = {}  # Dict[bytes: Appt]
        self._next_indx = 0  # index a new appt gets assigned
        self.tickoff = 0  # Used for test overrides

        self._wake_event = s_coro.Event()  # Causes the scheduler loop to wake up
        self.onfini(self._wake_event.set)

        self._hivenode = await self.core.hive.open(('agenda', 'appts'))  # Persistent storage
        self.onfini(self.stop)

        self.enabled = False
        self._schedtask = None  # The task of the scheduler loop.  Doesn't run until we're enabled

        self._running_tasks = []  # The actively running cron job tasks
        await self._load_all()

    async def start(self):
        '''
        Enable cron jobs to start running, start the scheduler loop

        Go through all the appointments, making sure the query is valid, and remove the ones that aren't.  (We can't
        evaluate queries until enabled because not all the modules are loaded yet.)
        '''
        if self.enabled:
            return

        await self._load_all()
        for iden, appt in self.appts.items():
            try:
                await self.core.getStormQuery(appt.query)
            except Exception as e:
                logger.exception(f'Invalid appointment {iden} {appt.name} found in storage. Disabling. {e}',
                                 extra={'synapse': {'iden': iden, 'name': appt.name, 'text': appt.query}})
                appt.enabled = False

        self._schedtask = self.schedCoro(self._scheduleLoop())
        self.enabled = True

    async def stop(self):
        "Cancel the scheduler loop, and set self.enabled to False."
        if not self.enabled:
            return
        self._schedtask.cancel()
        for task in self._running_tasks:
            await task.fini()

        self.enabled = False

    async def _load_all(self):
        '''
        Load all the appointments from persistent storage
        '''
        # Clear existing appointments before loading
        self.apptheap = []
        self.appts = {}

        to_delete = []
        for iden, node in iter(self._hivenode):
            val = node.valu
            try:
                appt = _Appt.unpack(self, val)
                if appt.iden != iden:
                    raise s_exc.InconsistentStorage(mesg='iden inconsistency')
                self._addappt(iden, appt)
                self._next_indx = max(self._next_indx, appt.indx + 1)
            except (s_exc.InconsistentStorage, s_exc.BadStorageVersion, s_exc.BadTime, TypeError, KeyError,
                    UnicodeDecodeError) as e:
                logger.warning('Invalid appointment %r found in storage: %r.  Removing.', iden, e)
                to_delete.append(iden)
                continue

        for iden in to_delete:
            node = self._hivenode.get(iden)
            if node is not None:
                await node.hive.pop(node.full)

        # Make sure we don't assign the same index to 2 appointments
        if self.appts:
            maxindx = max(appt.indx for appt in self.appts.values())
            self._next_indx = maxindx + 1

    def _addappt(self, iden, appt):
        '''
        Updates the data structures to add an appointment
        '''
        if appt.nexttime:
            heapq.heappush(self.apptheap, appt)
        self.appts[iden] = appt
        if self.apptheap and self.apptheap[0] is appt:
            self._wake_event.set()

    async def _storeAppt(self, appt, nexs=False):
        ''' Store a single appointment '''
        full = self._hivenode.full + (appt.iden,)
        stordict = appt.pack()

        # Don't store ephemeral props
        for prop in ('startcount', 'errcount', 'lasterrs'):
            stordict.pop(prop, None)

        await self.core.hive.set(full, stordict, nexs=nexs)

    @staticmethod
    def _dictproduct(rdict):
        '''
        Yields a series of dicts that cover the combination of all multiple-value (e.g. lists or tuples) values, with
        non-multiple-value values remaining the same.
        '''
        multkeys = [k for k, v in rdict.items() if isinstance(v, Iterable)]
        if not multkeys:
            yield rdict
            return

        multvals = [rdict[k] for k in multkeys]

        for combo in itertools.product(*multvals):
            newdict = rdict.copy()
            for i, k in enumerate(multkeys):
                newdict[k] = combo[i]
            yield newdict

    def list(self):
        return list(self.appts.items())

    async def add(self, cdef):
        '''
        Persistently adds an appointment

        Args:
            cdef (dict):  May contain the following keys:
                creator (str):  iden of the creating user
                iden (str): Iden of the appointment
                storm (str): storm query to run
                reqs (Union[None, Dict[TimeUnit, Union[int, Tuple[int]], List[...]):
                    one or more dicts of the fixed aspects of the appointment.  dict value may be a single or multiple.
                    May be an empty dict or None.
                incunit (Union[None, TimeUnit]):
                    the unit that changes for recurring, or None for non-recurring.  It is an error for this value to
                    match a key in reqdict.
                incvals (Union[None, int, Iterable[int]): count of units of incunit or explicit day of week or day of
                    month.  Not allowed for incunit == None, required for others (1 would be a typical value)

        Notes:
            For values in reqs that are lists and incvals if a list, all combinations of all values (the product) are
            used

        Returns:
            Packed appointment definition
        '''
        iden = cdef['iden']
        incunit = cdef.get('incunit')
        incvals = cdef.get('incvals')
        reqs = cdef.get('reqs', {})
        query = cdef.get('storm')
        creator = cdef.get('creator')
        view = cdef.get('view')

        recur = incunit is not None
        indx = self._next_indx
        self._next_indx += 1

        if iden in self.appts:
            raise s_exc.DupIden()

        if not query:
            raise ValueError('"query" key of cdef parameter is not present or empty')

        if not creator:
            raise ValueError('"creator" key is cdef parameter is not present or empty')

        if not reqs and incunit is None:
            raise ValueError('at least one of reqs and incunit must be non-empty')

        if incunit is not None and incvals is None:
            raise ValueError('incvals must be non-None if incunit is non-None')

        if isinstance(reqs, Mapping):
            reqs = [reqs]

        # Find all combinations of values in reqdict values and incvals values
        nexttime = None
        recs = []  # type: ignore
        for req in reqs:
            if TimeUnit.NOW in req:
                if incunit is not None:
                    mesg = "Recurring jobs may not be scheduled to run 'now'"
                    raise ValueError(mesg)
                nexttime = self._getNowTick()
                continue

            reqdicts = self._dictproduct(req)
            if not isinstance(incvals, Iterable):
                incvals = (incvals, )
            recs.extend(ApptRec(rd, incunit, v) for (rd, v) in itertools.product(reqdicts, incvals))

        appt = _Appt(self, iden, recur, indx, query, creator, recs, nexttime=nexttime, view=view)
        self._addappt(iden, appt)

        appt.doc = cdef.get('doc', '')

        await self._storeAppt(appt)

        return appt.pack()

    async def get(self, iden):

        appt = self.appts.get(iden)
        if appt is not None:
            return appt

        mesg = f'No cron job with id: {iden}'
        raise s_exc.NoSuchIden(iden=iden, mesg=mesg)

    async def enable(self, iden):
        appt = self.appts.get(iden)
        if appt is None:
            raise s_exc.NoSuchIden()

        await self.mod(iden, appt.query)

    async def disable(self, iden):
        appt = self.appts.get(iden)
        if appt is None:
            raise s_exc.NoSuchIden()

        appt.enabled = False
        await self._storeAppt(appt)

    async def mod(self, iden, query):
        '''
        Change the query of an appointment
        '''
        appt = self.appts.get(iden)
        if appt is None:
            raise s_exc.NoSuchIden()

        if not query:
            raise ValueError('empty query')

        if self.enabled:
            await self.core.getStormQuery(query)

        appt.query = query
        appt.enabled = True  # in case it was disabled for a bad query

        await self._storeAppt(appt)

    async def move(self, croniden, viewiden):
        '''
        Move a cronjob from one view to another
        '''
        appt = self.appts.get(croniden)
        if appt is None:
            raise s_exc.NoSuchIden()

        appt.view = viewiden

        await self._storeAppt(appt)

    async def delete(self, iden):
        '''
        Delete an appointment
        '''
        appt = self.appts.get(iden)
        if appt is None:
            raise s_exc.NoSuchIden()

        try:
            heappos = self.apptheap.index(appt)
        except ValueError:
            pass  # this is OK, just a non-recurring appt that has no more records
        else:
            # If we're already the last item, just delete it
            if heappos == len(self.apptheap) - 1:
                del self.apptheap[heappos]
            else:
                # put the last item at the current position and reheap
                self.apptheap[heappos] = self.apptheap.pop()
                heapq.heapify(self.apptheap)

        del self.appts[iden]
        node = self._hivenode.get(iden)
        if node is not None:
            await node.hive.pop(node.full)

    def _getNowTick(self):
        return time.time() + self.tickoff

    def _addTickOff(self, offs):
        self.tickoff += offs
        self._wake_event.set()

    async def _scheduleLoop(self):
        '''
        Task loop to issue query tasks at the right times.
        '''
        while True:

            timeout = None
            if self.apptheap:
                timeout = self.apptheap[0].nexttime - self._getNowTick()

            if timeout is None or timeout > 0:
                self._wake_event.clear()
                await self._wake_event.timewait(timeout=timeout)

            if self.isfini:
                return

            now = self._getNowTick()
            while self.apptheap and self.apptheap[0].nexttime <= now:

                appt = heapq.heappop(self.apptheap)
                appt.updateNexttime(now)

                if appt.nexttime:
                    heapq.heappush(self.apptheap, appt)

                if not appt.enabled or not self.enabled:
                    continue

                if appt.isrunning:  # pragma: no cover
                    mesg = f'Appointment {appt.iden} {appt.name} is still running from previous time when scheduled' \
                           f' to run. Skipping.',
                    logger.warning(mesg,
                                   extra={'synapse': {'iden': appt.iden, 'name': appt.name}})
                else:
                    await self._execute(appt)

    async def _execute(self, appt):
        '''
        Fire off the task to make the storm query
        '''
        user = self.core.auth.user(appt.creator)
        if user is None:
            logger.warning(f'Unknown user {appt.creator} in stored appointment {appt.iden} {appt.name}',
                           extra={'synapse': {'iden': appt.iden, 'name': appt.name, 'user': appt.creator}})
            await self._markfailed(appt, 'unknown user')
            return

        locked = user.info.get('locked')
        if locked:
            logger.warning(f'Cron {appt.iden} {appt.name} failed because creator {user.name} is locked',
                           extra={'synapse': {'iden': appt.iden, 'name': appt.name, 'user': appt.creator,
                                              'username': user.name}})
            await self._markfailed(appt, 'locked user')
            return

        view = self.core.getView(iden=appt.view, user=user)
        if view is None:
            logger.warning(f'Unknown view {appt.view} in stored appointment {appt.iden} {appt.name}',
                           extra={'synapse': {'iden': appt.iden, 'name': appt.name, 'user': appt.creator,
                                              'username': user.name, 'view': appt.view}})
            await self._markfailed(appt, 'unknown view')
            return

        info = {'iden': appt.iden, 'query': appt.query}
        task = await self.core.boss.execute(self._runJob(user, appt), f'Cron {appt.iden}', user, info=info)
        self._running_tasks.append(task)

        task.onfini(functools.partial(self._running_tasks.remove, task))

    async def _markfailed(self, appt, reason):
        appt.lastfinishtime = appt.laststarttime = self._getNowTick()
        appt.startcount += 1
        appt.isrunning = False
        appt.lastresult = f'Failed due to {reason}'
        if not self.isfini:
            await self._storeAppt(appt, nexs=True)

    async def _runJob(self, user, appt):
        '''
        Actually run the storm query, updating the appropriate statistics and results
        '''
        count = 0
        appt.isrunning = True
        appt.laststarttime = self._getNowTick()
        appt.startcount += 1
        await self._storeAppt(appt, nexs=True)

        with s_provenance.claim('cron', iden=appt.iden):
            logger.info(f'Agenda executing for iden={appt.iden}, name={appt.name} user={user.name}, view={appt.view}, query={appt.query}',
                        extra={'synapse': {'iden': appt.iden, 'name': appt.name, 'user': user.iden, 'text': appt.query,
                                           'username': user.name, 'view': appt.view}})
            starttime = self._getNowTick()
            success = False
            try:
                opts = {'user': user.iden, 'view': appt.view, 'vars': {'auto': {'iden': appt.iden, 'type': 'cron'}}}
                async for node in self.core.eval(appt.query, opts=opts):
                    count += 1
            except asyncio.CancelledError:
                result = 'cancelled'
                raise
            except Exception as e:
                result = f'raised exception {e}'
                logger.exception(f'Agenda job {appt.iden} {appt.name} raised exception',
                                 extra={'synapse': {'iden': appt.iden, 'name': appt.name}}
                                 )
            else:
                success = True
                result = f'finished successfully with {count} nodes'
            finally:
                finishtime = self._getNowTick()
                if not success:
                    appt.errcount += 1
                    appt.lasterrs.append(result)
                took = finishtime - starttime
                mesg = f'Agenda completed query for iden={appt.iden} name={appt.name} with result "{result}" ' \
                       f'took {took:.3f}s'
                logger.info(mesg, extra={'synapse': {'iden': appt.iden, 'name': appt.name, 'user': user.iden,
                                                     'result': result, 'username': user.name, 'took': took}})
                appt.lastfinishtime = finishtime
                appt.isrunning = False
                appt.lastresult = result
                if not self.isfini:
                    await self._storeAppt(appt, nexs=True)



import synapse.common as s_common
import synapse.cryotank as s_cryotank

import synapse.lib.const as s_const

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

logger = s_cryotank.logger

cryodata = (('foo', {'bar': 10}), ('baz', {'faz': 20}))

class CryoTest(s_t_utils.SynTest):

    async def test_cryo_cell_async(self):
        async with self.getTestCryo() as cryo:
            async with cryo.getLocalProxy() as prox:
                self.true(await prox.init('foo'))
                self.eq([], await alist(prox.rows('foo', 0, 1)))

    async def test_cryo_cell(self):
        with self.getTestDir() as dirn:
            async with self.getTestCryoAndProxy(dirn=dirn) as (cryo, prox):

                self.eq((), await prox.list())

                self.true(await prox.init('foo'))

                self.eq('foo', (await prox.list())[0][0])

                self.none(await prox.last('foo'))

                self.eq([], await alist(prox.rows('foo', 0, 1)))

                self.true(await prox.puts('foo', cryodata))

                info = await prox.list()
                self.eq('foo', info[0][0])
                self.eq(2, info[0][1].get('stat').get('entries'))

                self.true(await prox.puts('foo', cryodata))

                items = await alist(prox.slice('foo', 1, 3))
                self.eq(items[0][1][0], 'baz')

                metrics = await alist(prox.metrics('foo', 0, 9999))
                self.len(2, metrics)
                self.eq(2, metrics[0][1]['count'])

                self.eq(3, (await prox.last('foo'))[0])
                self.eq('baz', (await prox.last('foo'))[1][0])

                iden = s_common.guid()
                self.eq(0, await prox.offset('foo', iden))

                items = await alist(prox.slice('foo', 0, 1000, iden=iden))
                self.eq(0, await prox.offset('foo', iden))

                items = await alist(prox.slice('foo', 4, 1000, iden=iden))
                self.eq(4, await prox.offset('foo', iden))

                # test the direct tank share....
                async with cryo.getLocalProxy(share='cryotank/foo') as lprox:

                    items = await alist(lprox.slice(1, 3))

                    self.eq(items[0][1][0], 'baz')

                    self.len(4, await alist(lprox.slice(0, 9999)))

                    await lprox.puts(cryodata)

                    self.len(6, await alist(lprox.slice(0, 9999)))

                    # test offset storage and updating
                    iden = s_common.guid()
                    self.eq(0, await lprox.offset(iden))
                    self.eq(2, await lprox.puts(cryodata, seqn=(iden, 0)))
                    self.eq(2, await lprox.offset(iden))

                # test the new open share
                async with cryo.getLocalProxy(share='cryotank/lulz') as lprox:

                    self.len(0, await alist(lprox.slice(0, 9999)))

                    await lprox.puts(cryodata)

                    self.len(2, await alist(lprox.slice(0, 9999)))

                    self.len(1, await alist(lprox.metrics(0)))

                # Delete apis
                self.false(await prox.delete('newp'))
                self.true(await prox.delete('lulz'))

            # Re-open the tank and ensure that the deleted tank is not present.
            async with self.getTestCryoAndProxy(dirn=dirn) as (cryo, prox):
                tanks = await prox.list()
                self.len(1, tanks)
                self.eq('foo', tanks[0][0])

    async def test_cryo_init(self):
        with self.getTestDir() as dirn:
            async with self.getTestCryo(dirn) as cryo:
                # test passing conf data in through init directly
                tank = await cryo.init('conftest', conf={'map_size': s_const.mebibyte * 64})
                self.eq(tank.slab.mapsize, s_const.mebibyte * 64)
                _, conf = await cryo.hive.get(('cryo', 'names', 'conftest'))
                self.eq(conf, {'map_size': s_const.mebibyte * 64})

            # And the data was persisted
            async with self.getTestCryo(dirn) as cryo:
                tank = cryo.tanks.get('conftest')
                self.eq(tank.slab.mapsize, s_const.mebibyte * 64)
                _, conf = await cryo.hive.get(('cryo', 'names', 'conftest'))
                self.eq(conf, {'map_size': s_const.mebibyte * 64})



import os
import json
import asyncio
import synapse.exc as s_exc
import synapse.common as s_common

from synapse.common import aspin

import synapse.cortex as s_cortex
import synapse.telepath as s_telepath
import synapse.tests.utils as s_t_utils
import synapse.tools.backup as s_tools_backup

class TrigTest(s_t_utils.SynTest):

    async def test_trigger_async(self):

        with self.getTestDir() as dirn:

            async with self.getTestCore(dirn=dirn) as core:

                await core.stormlist('trigger.add node:add --async --form inet:ipv4 --query { [+#foo] $lib.queue.gen(foo).put($node.iden()) }')

                nodes = await core.nodes('[ inet:ipv4=1.2.3.4 ]')
                self.none(nodes[0].tags.get('foo'))

                msgs = await core.stormlist('trigger.list')
                self.stormIsInPrint('True   True   node:add  inet:ipv4', msgs)

                self.nn(await core.callStorm('return($lib.queue.gen(foo).pop(wait=$lib.true))'))
                nodes = await core.nodes('inet:ipv4=1.2.3.4')
                self.nn(nodes[0].tags.get('foo'))

                # test dynamically updating the trigger async to off
                await core.stormlist('$lib.view.get().triggers.0.set(async, $lib.false)')
                nodes = await core.nodes('[ inet:ipv4=5.5.5.5 ]')
                self.nn(nodes[0].tags.get('foo'))
                self.nn(await core.callStorm('return($lib.queue.gen(foo).pop(wait=$lib.true))'))

                # reset the trigger to async...
                await core.stormlist('$lib.view.get().triggers.0.set(async, $lib.true)')

                # kill off the async consumer and queue up some requests
                # to test persistance and proper resuming...
                await core.view.finiTrigTask()

                trigiden = await core.callStorm('return($lib.view.get().triggers.0.iden)')
                self.nn(trigiden)

                await core.view.addTrigQueue({'buid': s_common.buid(), 'trig': trigiden})
                await core.view.addTrigQueue({'buid': s_common.buid(), 'trig': s_common.guid()})

                nodes = await core.nodes('[ inet:ipv4=9.9.9.9 ]')
                self.none(nodes[0].tags.get('foo'))
                self.none(await core.callStorm('return($lib.queue.gen(foo).pop())'))

            async with self.getTestCore(dirn=dirn) as core:

                self.nn(await core.callStorm('return($lib.queue.gen(foo).pop(wait=$lib.true))'))
                nodes = await core.nodes('inet:ipv4=9.9.9.9')
                self.nn(nodes[0].tags.get('foo'))
                self.none(core.view.trigqueue.last())

                # lets fork a view and hamstring it's trigger queue and make sure we can't merge
                viewiden = await core.callStorm('return($lib.view.get().fork().iden)')

                view = core.getView(viewiden)
                await view.finiTrigTask()

                opts = {'view': viewiden}
                await core.stormlist('trigger.add node:add --async --form inet:ipv4 --query { [+#foo] $lib.queue.gen(foo).put($node.iden()) }', opts=opts)
                nodes = await core.nodes('[ inet:ipv4=123.123.123.123 ]', opts=opts)

                with self.raises(s_exc.CantMergeView):
                    await core.nodes('$lib.view.get().merge()', opts=opts)

                await core.nodes('$lib.view.get().merge(force=$lib.true)', opts=opts)
                await core.nodes('$lib.view.del($view)', opts={'vars': {'view': viewiden}})

                self.false(os.path.isdir(view.dirn))

    async def test_trigger_async_mirror(self):

        with self.getTestDir() as dirn:

            path00 = s_common.gendir(dirn, 'core00')
            path01 = s_common.gendir(dirn, 'core01')

            async with self.getTestCore(dirn=path00) as core00:
                await core00.stormlist('trigger.add node:add --async --form inet:ipv4 --query { [+#foo] $lib.queue.gen(foo).put($node.iden()) }')

                await core00.view.finiTrigTask()
                await core00.nodes('[ inet:ipv4=1.2.3.4 ]')

            s_tools_backup.backup(path00, path01)

            async with self.getTestCore(dirn=path00) as core00:

                url = core00.getLocalUrl()
                core01conf = {'mirror': url}

                async with await s_cortex.Cortex.anit(dirn=path01, conf=core01conf) as core01:
                    # ensure sync by forcing node construction
                    await core01.nodes('[ou:org=*]')
                    self.nn(await core00.callStorm('return($lib.queue.gen(foo).pop(wait=$lib.true))'))
                    self.none(await core00.callStorm('return($lib.queue.gen(foo).pop())'))

                    await core01.nodes('[inet:ipv4=8.8.8.8]')
                    self.nn(await core01.callStorm('return($lib.queue.gen(foo).pop(wait=$lib.true))'))
                    self.none(await core00.callStorm('return($lib.queue.gen(foo).pop())'))
                    self.none(await core01.callStorm('return($lib.queue.gen(foo).pop())'))

                    self.nn(core00.view.trigtask)
                    self.none(core01.view.trigtask)

    async def test_trigger_recursion(self):
        async with self.getTestCore() as core:
            tdef = {'cond': 'node:add', 'form': 'test:guid', 'storm': '[ test:guid="*" ]'}
            await core.view.addTrigger(tdef)
            await self.asyncraises(s_exc.RecursionLimitHit, core.nodes('[ test:guid="*" ]'))

    async def test_modification_persistence(self):

        with self.getTestDir() as fdir:

            async with self.getTestCore(dirn=fdir) as core:
                iden = s_common.guid()
                tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[inet:user=1] | testcmd'}
                await core.view.addTrigger(tdef)

                triggers = await core.view.listTriggers()
                self.eq(triggers[0][1].tdef['storm'], '[inet:user=1] | testcmd')

                iden = triggers[0][0]
                await core.view.setTriggerInfo(iden, 'storm', '[inet:user=2 .test:univ=4] | testcmd')
                triggers = await core.view.listTriggers()
                self.eq(triggers[0][1].tdef['storm'], '[inet:user=2 .test:univ=4] | testcmd')

                # Sad cases
                await self.asyncraises(s_exc.BadSyntax, core.view.setTriggerInfo(iden, 'storm', ' | | badstorm '))
                await self.asyncraises(s_exc.NoSuchIden, core.view.setTriggerInfo('deadb33f', 'storm', 'inet:user'))

    async def test_trigger_basics(self):

        async with self.getTestCoreAndProxy() as (core, prox):

            view = core.view

            # node:add case
            q = '''$s=$lib.str.format("f={f} v={v}", f=$auto.opts.form, v=$auto.opts.valu) $lib.log.info($s)
                    [ test:guid="*" +#nodeadd]'''
            tdef = {'cond': 'node:add', 'form': 'test:str', 'storm': q}
            await view.addTrigger(tdef)
            with self.getAsyncLoggerStream('synapse.storm.log', 'f=') as stream:
                await core.nodes('[ test:str=foo ]')
                self.true(await stream.wait(6))
            self.eq(stream.getvalue().strip(), 'f=test:str v=foo')
            self.len(1, await core.nodes('test:guid#nodeadd'))

            # node:del case
            tdef = {'cond': 'node:del', 'storm': '[ test:guid="*" +#nodedel ]', 'form': 'test:str'}
            await view.addTrigger(tdef)
            await core.nodes('test:str=foo | delnode')
            self.len(1, await core.nodes('test:guid#nodedel'))

            # tag:add case
            tdef = {'cond': 'tag:add', 'storm': '[ test:guid="*" +#tagadd]', 'tag': 'footag'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#footag ]')
            self.len(1, await core.nodes('test:guid#tagadd'))

            # tag:add globbing and storm var
            tdef = {'cond': 'tag:add',
                    'storm': '$lib.log.info($auto.opts.tag) [ +#count test:str=$tag ]',
                    'tag': 'a.*.c'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#a.b ]')
            with self.getAsyncLoggerStream('synapse.storm.log', 'a.b.c') as stream:
                await core.nodes('[ test:str=foo +#a.b.c ]')
                self.true(await stream.wait(6))
            self.true(stream.getvalue().strip().startswith('a.b.c\n'))
            await core.nodes('[ test:str=foo +#a.b.ccc ]')
            self.len(1, await core.nodes('#count'))
            self.len(1, await core.nodes('test:str=a.b.c'))

            tdef = {'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo.**.baz'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#foo.1.2.3.baz ]')
            self.len(1, await core.nodes('test:str=foo.1.2.3.baz'))

            # tag:del case
            tdef = {'cond': 'tag:del', 'storm': '[ test:guid="*" +#tagdel ]', 'tag': 'footag'}
            await view.addTrigger(tdef)
            await core.nodes('test:str=foo [ -#footag ]')
            self.len(1, await core.nodes('test:guid#tagdel'))

            # Form/tag add
            tdef = {'cond': 'tag:add', 'storm': '[ test:guid="*" +#formtagadd]', 'tag': 'bartag',
                    'form': 'test:str'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=foo +#bartag ]')
            self.len(1, await core.nodes('test:guid#formtagadd'))

            # Wrong form/right tag add doesn't fire
            await core.nodes('test:int=5 | delnode')
            self.len(0, await core.nodes('test:int=5'))
            await core.nodes('[ test:auto=1 +#bartag ]')
            self.len(0, await core.nodes('test:int=5'))

            # Right form/wrong tag add doesn't fire
            await core.nodes('[ test:str=bar +#footag ]')
            self.len(0, await core.nodes('test:int=5'))

            # Prop set
            q = '''$s=$lib.str.format("pf={f} pn={n}", f=$auto.opts.propfull, n=$auto.opts.propname) $lib.log.info($s)
            [ test:guid="*" +#propset ]'''
            tdef = {'cond': 'prop:set',
                    'storm': q,
                    'prop': 'test:type10:intprop'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:type10=1 ]')
            with self.getAsyncLoggerStream('synapse.storm.log', 'pf=') as stream:
                await core.nodes('[ test:type10=1 :intprop=25 ]')
                self.true(await stream.wait(6))
            buf = stream.getvalue().strip()
            self.eq(buf, 'pf=test:type10:intprop pn=intprop')
            self.len(1, await core.nodes('test:guid#propset'))

            # Test re-setting doesn't fire
            await core.nodes('test:int=6 | delnode')
            await core.nodes('[ test:type10=1 :intprop=25 ]')
            self.len(0, await core.nodes('test:int=6'))

            # Prop set univ
            tdef = {'cond': 'prop:set', 'storm': '[ test:guid="*" +#propsetuniv ]', 'prop': '.test:univ'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:type10=1 .test:univ=1 ]')
            self.len(1, await core.nodes('test:guid#propsetuniv'))

            # Prop set form specific univ
            tdef = {'cond': 'prop:set', 'storm': '[ test:guid="*" +#propsetuniv2 ]', 'prop': 'test:str.test:univ'}
            await view.addTrigger(tdef)
            await core.nodes('[ test:str=beep .test:univ=1 ]')
            self.len(1, await core.nodes('test:guid#propsetuniv2'))

            # Add trigger with iden
            iden = s_common.guid()
            tdef0 = {'cond': 'node:add', 'storm': '[ +#withiden ]', 'form': 'test:int', 'iden': iden}
            await view.addTrigger(tdef0)
            self.nn(await view.getTrigger(iden))
            await core.nodes('[ test:int=77 ]')
            self.len(1, await core.nodes('test:int#withiden'))

            # iden embedded in vars
            q = '+test:str~=log $s=$lib.str.format("test {t} {i}", t=$auto.type, i=$auto.iden) $lib.log.info($s, ({"iden": $auto.iden}))'
            tdef = {'cond': 'node:add', 'form': 'test:str', 'storm': q}
            await view.addTrigger(tdef)
            with self.getStructuredAsyncLoggerStream('synapse.storm.log', 'test trigger') as stream:
                await core.nodes('[ test:str=logit ]')
                self.true(await stream.wait(6))
            buf = stream.getvalue()
            mesg = json.loads(buf.split('\n')[-2])
            self.eq(mesg['message'], f'test trigger {tdef.get("iden")}')
            self.eq(mesg['iden'], tdef.get('iden'))

            # Attempting to add trigger with existing iden raises
            with self.raises(s_exc.DupIden):
                tdef = {'cond': 'node:add', 'storm': '[ +#dupiden ]', 'form': 'test:int', 'iden': iden}
                await view.addTrigger(tdef)
            self.eq(tdef0, (await view.getTrigger(iden)).tdef)

            # Bad trigger parms
            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'nocond', 'storm': 'test:int=4', 'form': 'test:str'})

            with self.raises(s_exc.BadSyntax):
                await view.addTrigger({'cond': 'node:add', 'storm': ' | | badstorm ', 'form': 'test:str'})

            with self.raises(s_exc.BadOptValu):
                await view.addTrigger({'cond': 'node:add', 'storm': 'test:int=4', 'form': 'test:str', 'tag': 'foo'})

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'prop:set', 'storm': 'test:int=4', 'form': 'test:str', 'prop': 'foo'})

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]'})

            with self.raises(s_exc.BadOptValu):
                tdef = {'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo', 'prop': 'test:str'}
                await view.addTrigger(tdef)

            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'node:add', 'storm': 'test:int=4', 'form': 'test:str', 'iden': 'foo'})

            # bad tagmatch
            with self.raises(s_exc.BadConfValu):
                await view.addTrigger({'cond': 'tag:add', 'storm': '[ +#count test:str=$tag ]', 'tag': 'foo&baz'})

            # Trigger list
            triglist = await view.listTriggers()
            self.len(12, triglist)

            # Delete not a trigger
            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger('foo'))

            # Delete trigger
            iden = [iden for iden, r in triglist if r.tdef['cond'] == 'prop:set'][0]
            await core.nodes('test:int=6 | delnode')
            await view.delTrigger(iden)

            # The deleted trigger shall not fire
            await core.nodes('[ test:type10=3 :intprop=25 ]')
            self.len(0, await core.nodes('test:int=6'))

            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger(b'badiden'))

            # Mod trigger

            iden = [iden for iden, r in triglist if r.tdef['cond'] == 'tag:add' and r.tdef.get('form') == 'test:str'][0]

            await view.setTriggerInfo(iden, 'storm', '[ test:int=42 ]')
            await core.nodes('[ test:str=foo4 +#bartag ]')
            self.len(1, await core.nodes('test:int=42'))

            # Delete a tag:add
            iden2 = [iden for iden, r in triglist if r.tdef['cond'] == 'tag:add' and r.tdef.get('form') is None][0]
            await view.delTrigger(iden2)

            # A rando user can't manipulate triggers

            user = await core.auth.addUser('fred')
            await user.setPasswd('fred')

            url = core.getLocalUrl()

            async with await s_telepath.openurl(url, user='fred') as fred:
                # Trigger list other user
                self.len(0, await fred.callStorm('return ($lib.trigger.list())'))

                with self.raises(s_exc.AuthDeny):
                    tdef = {'cond': 'node:add', 'storm': '[ test:int=1 ]', 'form': 'test:str'}
                    opts = {'vars': {'tdef': tdef}}
                    q = 'return ($lib.trigger.add($tdef))'
                    await fred.callStorm(q, opts=opts)

                # Delete trigger auth failure
                await self.asyncraises(s_exc.StormRuntimeError, fred.callStorm(f'$lib.trigger.del({iden})'))

                # Mod trigger auth failure
                opts = {'vars': {'iden': iden}}
                await self.asyncraises(s_exc.StormRuntimeError,
                                       fred.callStorm('$lib.trigger.mod($iden, "{#foo}")', opts=opts))

            # additional NoSuchIden failures
            await self.asyncraises(s_exc.NoSuchIden, view.getTrigger('newp'))
            await self.asyncraises(s_exc.NoSuchIden, view.delTrigger('newp'))
            await self.asyncraises(s_exc.NoSuchIden, view.setTriggerInfo('newp', 'enabled', True))

    async def test_trigger_delete(self):

        async with self.getTestCore() as core:

            tdef0 = {'cond': 'node:add', 'form': 'test:guid', 'storm': '[test:str=add]'}
            tdef1 = {'cond': 'node:del', 'form': 'test:guid', 'storm': '[test:str=del]'}
            tdef2 = {'cond': 'prop:set', 'prop': 'test:guid:tick', 'storm': '[test:str=set]'}

            tdef0 = await core.view.addTrigger(tdef0)
            tdef1 = await core.view.addTrigger(tdef1)
            tdef2 = await core.view.addTrigger(tdef2)

            await core.nodes('[test:guid="*" :tick=2015] | delnode')
            self.len(3, await core.nodes('test:str'))

            await core.view.delTrigger(tdef0.get('iden'))
            await core.view.delTrigger(tdef1.get('iden'))
            await core.view.delTrigger(tdef2.get('iden'))

            await core.nodes('test:str | delnode')
            await core.nodes('[test:guid="*" :tick=2015] | delnode')

            self.len(0, await core.nodes('test:str'))

    async def test_trigger_tag_globs(self):

        async with self.getTestCore() as core:

            root = await core.auth.getUserByName('root')

            tdef = {'iden': '1', 'user': root.iden, 'cond': 'tag:add', 'storm': '$lib.queue.get(foo).put(count0)',
                    'tag': 'foo.*.bar', 'enabled': True}
            trig1 = await core.view.triggers.load(tdef)

            tdef = {'iden': '2', 'user': root.iden, 'cond': 'tag:del', 'storm': '$lib.queue.get(foo).put(count1)',
                    'tag': 'baz.*.faz', 'form': 'test:guid', 'enabled': True}
            trig2 = await core.view.triggers.load(tdef)

            await core.nodes('$lib.queue.add(foo)')

            async def popNextFoo():
                return await core.callStorm('''
                    return ($lib.queue.get(foo).pop().index(1))
                ''')

            await core.nodes('[ test:guid="*" +#foo.asdf.bar ]')
            await core.nodes('[ test:guid="*" +#baz.asdf.faz ]')
            await core.nodes('#baz.asdf.faz [ -#baz.asdf.faz ]')

            self.eq('count0', await popNextFoo())
            self.eq('count1', await popNextFoo())

            core.view.triggers.pop(trig1.iden)
            core.view.triggers.pop(trig2.iden)

            await core.nodes('test:guid | delnode')

            # Triggers don't fire after they've been deleted

            await core.nodes('[ test:guid="*" +#foo.asdf.bar ]')
            await core.nodes('[ test:guid="*" +#baz.asdf.faz ]')

            await core.nodes('#baz.asdf.faz [ -#baz.asdf.faz ]')

            self.eq(0, await core.callStorm('return ($lib.queue.get(foo).size())'))

    async def test_trigger_running_perms(self):
        async with self.getTestCore() as core:
            visi = await core.auth.addUser('visi')

            # Verify that triggers are run as user
            tdef = {'cond': 'node:add', 'form': 'test:str', 'storm': '[ test:guid="*" +#nodeadd]', 'user': visi.iden}
            await core.view.addTrigger(tdef)
            await core.nodes('[ test:str=foo ]')
            self.len(0, await core.nodes('test:guid#nodeadd'))

            await visi.addRule((True, ('node', 'tag', 'add')))
            await visi.addRule((True, ('node', 'add')))
            await core.nodes('[ test:str=foo2 ]')
            self.len(1, await core.nodes('test:guid#nodeadd'))

            # If the user is locked, the trigger doesn't run
            await visi.setLocked(True)
            await core.nodes('[ test:str=foo3 ]')
            self.len(1, await core.nodes('test:guid#nodeadd'))

    async def test_trigger_perms(self):

        async with self.getTestCore() as core:

            visi = await core.auth.addUser('visi')
            newb = await core.auth.addUser('newb')

            await visi.addRule((True, ('node', 'tag', 'add', 'foo')))

            async with core.getLocalProxy(user='visi') as proxy:

                with self.raises(s_exc.AuthDeny):
                    tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[ +#foo ]'}
                    await proxy.callStorm('return ($lib.trigger.add($tdef).get(iden))', opts={'vars': {'tdef': tdef}})

                await visi.addRule((True, ('trigger', 'add')))

                tdef = {'cond': 'node:add', 'form': 'inet:ipv4', 'storm': '[ +#foo ]'}
                opts = {'vars': {'tdef': tdef}}
                trig = await proxy.callStorm('return ($lib.trigger.add($tdef))', opts=opts)
                iden0 = trig['iden']

                iden1 = s_common.guid()
                tdef = {'cond': 'node:add', 'form': 'inet:ipv6', 'storm': '[ +#foo ]', 'iden': iden1}
                opts = {'vars': {'tdef': tdef}}
                trig = await proxy.callStorm('return ($lib.trigger.add($tdef))', opts=opts)

                nodes = await core.nodes('[ inet:ipv4=1.2.3.4 ]')
                self.nn(nodes[0].tags.get('foo'))

                await aspin(proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': iden1}}))

            trigs = await core.view.listTriggers()
            trigiden = trigs[0][0]
            self.eq(trigiden, iden0)

            async with core.getLocalProxy(user='newb') as proxy:

                await self.agenlen(1, proxy.eval('syn:trigger'))

                await newb.addRule((True, ('trigger', 'get')))
                with self.raises(s_exc.AuthDeny):
                    await proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': trigs[0][0]}}).list()

                await self.agenlen(1, proxy.eval('syn:trigger'))

                with self.raises(s_exc.AuthDeny):
                    opts = {'vars': {'iden': trigiden}}
                    await proxy.eval('$lib.trigger.get($iden).set(enabled, $(0))', opts=opts).list()

                await newb.addRule((True, ('trigger', 'set')))
                opts = {'vars': {'iden': trigiden}}
                await aspin(proxy.eval('$lib.trigger.get($iden).set(enabled, $(0))', opts=opts))

                await newb.addRule((True, ('trigger', 'del')))
                await aspin(proxy.eval('$lib.trigger.del($iden)', opts={'vars': {'iden': trigiden}}))

            # If the trigger owner loses read perms on the trigger's view, it doesn't fire.
            # Regression test:  it also doesn't stop the pipeline/raise an exception
            await visi.addRule((False, ('view', 'read')), gateiden=core.view.iden)
            await newb.addRule((True, ('node', 'add')))
            async with core.getLocalProxy(user='newb') as proxy:
                self.eq(0, await proxy.count('[inet:ipv4 = 99] +#foo'))

    async def test_trigger_runts(self):

        async with self.getTestCore() as core:

            tdef = await core.view.addTrigger({
                'cond': 'node:add',
                'form': 'test:str',
                'storm': '[ test:int=1 ]',
            })
            iden = tdef['iden']

            nodes = await core.nodes('syn:trigger')
            self.len(1, nodes)
            self.eq(nodes[0].get('doc'), '')

            nodes = await core.nodes(f'syn:trigger={iden} [ :doc="hehe haha" :name=visitrig ]')
            self.eq(nodes[0].get('doc'), 'hehe haha')
            self.eq(nodes[0].get('name'), 'visitrig')

            nodes = await core.nodes(f'syn:trigger={iden}')
            self.eq(nodes[0].get('doc'), 'hehe haha')
            self.eq(nodes[0].get('name'), 'visitrig')



import logging

import synapse.exc as s_exc

import synapse.lib.module as s_module

logger = logging.getLogger(__name__)

class SynModule(s_module.CoreModule):

    def initCoreModule(self):

        for form, lifter in (('syn:cmd', self._liftRuntSynCmd),
                             ('syn:cron', self._liftRuntSynCron),
                             ('syn:form', self._liftRuntSynForm),
                             ('syn:prop', self._liftRuntSynProp),
                             ('syn:type', self._liftRuntSynType),
                             ('syn:tagprop', self._liftRuntSynTagProp),
                             ('syn:trigger', self._liftRuntSynTrigger),
                             ):
            form = self.model.form(form)
            self.core.addRuntLift(form.full, lifter)
            for _, prop in form.props.items():
                pfull = prop.full
                self.core.addRuntLift(pfull, lifter)

    async def _liftRuntSynCmd(self, full, valu=None, cmpr=None, view=None):

        genr = self.core.stormcmds.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynCron(self, full, valu=None, cmpr=None, view=None):

        genr = self.core.agenda.appts.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynForm(self, full, valu=None, cmpr=None, view=None):

        genr = self.model.forms.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynProp(self, full, valu=None, cmpr=None, view=None):

        genr = self.model.getProps

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynType(self, full, valu=None, cmpr=None, view=None):

        genr = self.model.types.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynTagProp(self, full, valu=None, cmpr=None, view=None):

        genr = self.model.tagprops.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _liftRuntSynTrigger(self, full, valu=None, cmpr=None, view=None):

        view = self.core.getView(iden=view)
        genr = view.triggers.triggers.values

        async for node in self._doRuntLift(genr, full, valu, cmpr):
            yield node

    async def _doRuntLift(self, genr, full, valu=None, cmpr=None):

        if cmpr is not None:
            filt = self.model.prop(full).type.getCmprCtor(cmpr)(valu)
            if filt is None:
                raise s_exc.BadCmprValu(cmpr=cmpr)

        fullprop = self.model.prop(full)
        if fullprop.isform:

            if cmpr is None:
                for obj in genr():
                    yield obj.getStorNode(fullprop)
                return

            for obj in genr():
                sode = obj.getStorNode(fullprop)
                if filt(sode[1]['ndef'][1]):
                    yield sode
        else:
            for obj in genr():
                sode = obj.getStorNode(fullprop.form)
                propval = sode[1]['props'].get(fullprop.name)

                if propval is not None and (cmpr is None or filt(propval)):
                    yield sode

    def getModelDefs(self):

        return (('syn', {

            'types': (
                ('syn:type', ('str', {'strip': True}), {
                    'doc': 'A Synapse type used for normalizing nodes and properties.',
                }),
                ('syn:form', ('str', {'strip': True}), {
                    'doc': 'A Synapse form used for representing nodes in the graph.',
                }),
                ('syn:prop', ('str', {'strip': True}), {
                    'doc': 'A Synapse property.'
                }),
                ('syn:tagprop', ('str', {'strip': True}), {
                    'doc': 'A user defined tag property.'
                }),
                ('syn:cron', ('guid', {}), {
                    'doc': 'A Cortex cron job.',
                }),
                ('syn:trigger', ('guid', {}), {
                    'doc': 'A Cortex trigger.'
                }),
                ('syn:cmd', ('str', {'strip': True}), {
                    'doc': 'A Synapse storm command.'
                }),
                ('syn:splice', ('guid', {'strip': True}), {
                    'doc': 'A splice from a layer.'
                }),
                ('syn:nodedata', ('comp', {'fields': (('key', 'str'), ('form', 'syn:form'))}), {
                    'doc': 'A nodedata key and the form it may be present on.',
                }),
                ('syn:user', ('guid', {'strip': True}), {
                    'doc': 'A Synapse user GUID.'
                }),
                ('syn:role', ('guid', {'strip': True}), {
                    'doc': 'A Synapse role GUID.'
                }),
            ),

            'forms': (

                ('syn:tag', {}, (

                    ('up', ('syn:tag', {}), {'ro': True,
                        'doc': 'The parent tag for the tag.'}),

                    ('isnow', ('syn:tag', {}), {
                        'doc': 'Set to an updated tag if the tag has been renamed.'}),

                    ('doc', ('str', {}), {
                        'doc': 'A short definition for the tag.',
                        'disp': {'hint': 'text'},
                    }),

                    ('doc:url', ('inet:url', {}), {
                        'doc': 'A URL link to additional documentation about the tag.'}),

                    ('depth', ('int', {}), {'ro': True,
                        'doc': 'How deep the tag is in the hierarchy.'}),

                    ('title', ('str', {}), {'doc': 'A display title for the tag.'}),

                    ('base', ('str', {}), {'ro': True,
                        'doc': 'The tag base name. Eg baz for foo.bar.baz .'}),
                )),
                ('syn:type', {'runt': True}, (
                    ('doc', ('str', {'strip': True}), {
                        'doc': 'The docstring for the type.', 'ro': True}),
                    ('ctor', ('str', {'strip': True}), {
                        'doc': 'The python ctor path for the type object.', 'ro': True}),
                    ('subof', ('syn:type', {}), {
                        'doc': 'Type which this inherits from.', 'ro': True}),
                    ('opts', ('data', {}), {
                        'doc': 'Arbitrary type options.', 'ro': True})
                )),
                ('syn:form', {'runt': True}, (
                    ('doc', ('str', {'strip': True}), {
                        'doc': 'The docstring for the form.', 'ro': True}),
                    ('type', ('syn:type', {}), {
                        'doc': 'Synapse type for this form.', 'ro': True}),
                    ('runt', ('bool', {}), {
                        'doc': 'Whether or not the form is runtime only.', 'ro': True})
                )),
                ('syn:prop', {'runt': True}, (
                    ('doc', ('str', {'strip': True}), {
                        'doc': 'Description of the property definition.'}),
                    ('form', ('syn:form', {}), {
                        'doc': 'The form of the property.', 'ro': True}),
                    ('type', ('syn:type', {}), {
                        'doc': 'The synapse type for this property.', 'ro': True}),
                    ('relname', ('str', {'strip': True}), {
                        'doc': 'Relative property name.', 'ro': True}),
                    ('univ', ('bool', {}), {
                        'doc': 'Specifies if a prop is universal.', 'ro': True}),
                    ('base', ('str', {'strip': True}), {
                        'doc': 'Base name of the property.', 'ro': True}),
                    ('ro', ('bool', {}), {
                        'doc': 'If the property is read-only after being set.', 'ro': True}),
                    ('extmodel', ('bool', {}), {
                        'doc': 'If the property is an extended model property or not.', 'ro': True}),
                )),
                ('syn:tagprop', {'runt': True}, (
                    ('doc', ('str', {'strip': True}), {
                        'doc': 'Description of the tagprop definition.'}),
                    ('type', ('syn:type', {}), {
                        'doc': 'The synapse type for this tagprop.', 'ro': True}),
                )),
                ('syn:trigger', {'runt': True}, (
                    ('vers', ('int', {}), {
                        'doc': 'Trigger version.', 'ro': True,
                    }),
                    ('doc', ('str', {}), {
                        'doc': 'A documentation string describing the trigger.',
                        'disp': {'hint': 'text'},
                    }),
                    ('name', ('str', {}), {
                        'doc': 'A user friendly name/alias for the trigger.',
                    }),
                    ('cond', ('str', {'strip': True, 'lower': True}), {
                        'doc': 'The trigger condition.', 'ro': True,
                    }),
                    ('user', ('str', {}), {
                        'doc': 'User who owns the trigger.', 'ro': True,
                    }),
                    ('storm', ('str', {}), {
                        'doc': 'The Storm query for the trigger.', 'ro': True,
                        'disp': {'hint': 'text'},
                    }),
                    ('enabled', ('bool', {}), {
                        'doc': 'Trigger enabled status.', 'ro': True,
                    }),
                    ('form', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'Form the trigger is watching for.'
                    }),
                    ('prop', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'Property the trigger is watching for.'
                    }),
                    ('tag', ('str', {'lower': True, 'strip': True}), {
                        'doc': 'Tag the trigger is watching for.'
                    }),
                )),
                ('syn:cron', {'runt': True}, (

                    ('doc', ('str', {}), {
                        'doc': 'A description of the cron job.',
                        'disp': {'hint': 'text'},
                    }),

                    ('name', ('str', {}), {
                        'doc': 'A user friendly name/alias for the cron job.'}),

                    ('storm', ('str', {}), {
                        'ro': True,
                        'doc': 'The storm query executed by the cron job.',
                        'disp': {'hint': 'text'},
                    }),

                )),
                ('syn:cmd', {'runt': True}, (
                    ('doc', ('str', {'strip': True}), {
                        'doc': 'Description of the command.',
                        'disp': {'hint': 'text'},
                    }),
                    ('package', ('str', {'strip': True}), {
                        'doc': 'Storm package which provided the command.'}),
                    ('svciden', ('guid', {'strip': True}), {
                        'doc': 'Storm service iden which provided the package.'}),
                    ('input', ('array', {'type': 'syn:form'}), {
                        'doc': 'The list of forms accepted by the command as input.', 'uniq': True, 'sorted': True, 'ro': True}),
                    ('output', ('array', {'type': 'syn:form'}), {
                        'doc': 'The list of forms produced by the command as output.', 'uniq': True, 'sorted': True, 'ro': True}),
                    ('nodedata', ('array', {'type': 'syn:nodedata'}), {
                        'doc': 'The list of nodedata that may be added by the command.', 'uniq': True, 'sorted': True, 'ro': True}),
                )),
                ('syn:splice', {'runt': True}, (
                    ('type', ('str', {'strip': True}), {
                        'doc': 'Type of splice.', 'ro': True
                    }),
                    ('iden', ('str', {}), {
                        'doc': 'The iden of the node involved in the splice.', 'ro': True,
                    }),
                    ('form', ('syn:form', {'strip': True}), {
                        'doc': 'The form involved in the splice.', 'ro': True
                    }),
                    ('prop', ('syn:prop', {'strip': True}), {
                        'doc': 'Property modified in the splice.', 'ro': True
                    }),
                    ('tag', ('syn:tag', {'strip': True}), {
                        'doc': 'Tag modified in the splice.', 'ro': True
                    }),
                    ('valu', ('data', {}), {
                        'doc': 'The value being set in the splice.', 'ro': True
                    }),
                    ('oldv', ('data', {}), {
                        'doc': 'The value before the splice.', 'ro': True
                    }),
                    ('user', ('guid', {}), {
                        'doc': 'The user who caused the splice.', 'ro': True,
                    }),
                    ('prov', ('guid', {}), {
                        'doc': 'The provenance stack of the splice.', 'ro': True,
                    }),
                    ('time', ('time', {}), {
                        'doc': 'The time the splice occurred.', 'ro': True,
                    }),
                    ('splice', ('str', {'strip': True}), {
                        'doc': 'The splice.', 'ro': True
                    }),
                )),
            ),
        }),)
